import ipyvuetify as v

from plotly.express.colors import qualitative

from seeq_sysid.gui.transfer_chip import TransferChip
from seeq_sysid.model.tf_item import TransferItem, TransferOption
from seeq_sysid.model.tf import TF

from pandas import DataFrame, concat


class TransferMatrix(v.Card):
    def __init__(self,
                 df=DataFrame(),
                 
                 elevation=0,
                 **kwargs):
        super().__init__(elevation=elevation,
                         style_='width:100%; overflow:auto auto',
                         **kwargs)
        self.mv = None
        self.cv = None
        self.matrix_dic = {}
        self.cv_label_dict = {}
        self.card_colors = qualitative.Plotly
        self.df = df.copy()
        self.model = None
        
        self.temp_df = None
        self.train_df = DataFrame()
        self.validation_df = DataFrame()
        
        self.signal_df = None
        self.capsules_df = None
        
        TransferChip.reset()

    def create_matrix(self, signal_df=DataFrame(), capsule_df=DataFrame(), mv=[], cv=[], train_capsules=[], valid_capsules=[]):
        self.signal_df = signal_df.copy()
        self.capsule_df = capsule_df.copy()
        self.mv = mv
        self.cv = cv
        matrix_list = []
        header_list = []
        row_list = []
        rows_list = []
        n_cv = 0
        n_mv = len(self.mv)
        
        # Create Header
        blank_block = v.Card(children=[''], class_='d-flex flex-column ma-1 pa-0 justify-center align-center', style_='width:170px; height:40px; font-weight:bold; font-size:10pt')
        header_list.append(blank_block)
        for mv_i in self.mv:
            header_list.append(v.Card(children=[mv_i], class_='d-flex flex-row ma-1 pa-0 justify-center align-center', style_='width:168px; height:40px; font-weight:bold; font-size:10pt'))
        
        blank_block1 = v.Card(children=[''], class_='d-flex flex-column ma-1 pa-0 justify-center align-center', elevation=0, style_='width:170px; height:40px; font-weight:bold; font-size:10pt')
        mv_label = [v.Card(children=[blank_block1, v.Card(children=['MV'], class_='d-flex flex-column ma-1 mb-0 pa-0 justify-center align-center', 
                                                        style_='width:{}px; height:40px; font-weight:bold; font-size:10pt'.format(174*n_mv))], 
                          class_='d-flex justify-left align-left ma-1 pa-0',
                          style_='font-weight:bold; font-size:10pt',
                          elevation=0)]
        
        matrix_list.append(v.Card(children=mv_label, style_='', class_='d-flex justify-left align-left flex-row ma-1 pa-0 ml-0 mb-0', elevation=0))
        matrix_list.append(v.Card(children=header_list, style_='', class_='d-flex justify-left flex-row ma-1 mt-0', elevation=0))

        
        for cv_i in self.cv:
            cv_label = CVLabel(cv_i=cv_i, v_on = cv_i+'capsulemenu.on')

            self.cv_label_dict[cv_i] = CapsulesCard(items=capsule_df.columns.to_list(), selected_train_conditions=train_capsules, selected_valid_conditions=valid_capsules)
            capsules_menu = v.Menu(value=None, children=[self.cv_label_dict[cv_i]], class_='ma-2 pa-2', close_on_content_click=False, v_slots=[{
                             'name': 'activator',
                             'variable': cv_i+'capsulemenu',
                             'children': [cv_label]
                         }],)
            
            
            self.matrix_dic[cv_i] = {}
            row_list = [cv_label, capsules_menu]

            for mv_i in self.mv:
                self.matrix_dic[cv_i][mv_i] = TransferChip(mv_i, cv_i)
                self.matrix_dic[cv_i][mv_i].update_meas_figure(self.signal_df)
            
            row_list.extend(list(self.matrix_dic[cv_i].values()))
            matrix_list.append(v.Card(children=row_list, style_='', class_='d-flex justify-left flex-row ma-1', elevation=0))
            
            n_cv += 1
        
        self.children = matrix_list
    
    def run(self, *args):
        self.to_identify()
        
    def compile_matrix(self):
        for cv_i in self.cv:
            for mv_i in self.mv:
                self.matrix_dic[cv_i][mv_i].compile_item()
                
    def create_model(self):
        self.model = TF()
        self.model.cv = []
        
        self.compile_matrix()

        for cv_i in self.cv:            
            transfer_item = TransferItem()
            transfer_item.cv = [cv_i]
            
            TransferOption.reset_idx()

            for mv_i in self.mv:
                item: TransferChip = self.matrix_dic[cv_i][mv_i]
                if item.tf_card_btn.value:
                    item.tf_card_btn.children = [item.loading_settings]
                item.tf_card_btn.disabled = True
                
                if self.matrix_dic[cv_i][mv_i].tf_card_btn.value:    
                    order = item.order
                    no_gain = item.no_gain
                    no_ramp = item.no_ramp
                    is_dt = item.is_dt
                    
                    gain_lb = item.gain_lb
                    gain_ub = item.gain_ub
                    
                    tau_lb = item.tau_lb
                    tau_ub = item.tau_ub
                    
                    deadtime_lb = item.deadtime_lb
                    deadtime_ub = item.deadtime_ub
                    
                    tf_option = TransferOption(order=order,
                                               no_gain=no_gain,
                                               no_ramp=no_ramp,
                                               is_dt=is_dt,
                                               gain_lb=gain_lb,
                                               gain_ub=gain_ub,
                                               tau_lb=tau_lb,
                                               tau_ub=tau_ub,
                                               theta_lb=deadtime_lb,
                                               theta_ub=deadtime_ub,
                                               name=mv_i)
                    
                    tf_option.id = TransferOption.idx
                    transfer_item.option_dict[mv_i] = tf_option
                    TransferOption.add_idx()
                
            if TransferOption.idx:
                self.model.add_model(cv_i, transfer_item)
    
    
    def create_dataset(self, capsules: list = None):
        signal_df = self.signal_df.copy()
        capsule_df = self.capsule_df.copy()

        if not capsules:
            return signal_df

        return signal_df[capsule_df[capsules].sum(axis=1) == True]
    
    
    def to_identify(self):
        signal_df = self.signal_df.copy()
        self.create_model()
        
        self.train_df = DataFrame()
        
        for cv_i in self.model.cv:
            # try:
            capsules = self.cv_label_dict[cv_i].select_train_capsules.v_model
            train_df = self.create_dataset(capsules)
            y_train_df = self.model.identify(train_df, cv_i)

            self.train_df = concat([self.train_df, y_train_df], axis=1)
            self.to_step_response(cv_i)

            self.train_df = concat([self.train_df, train_df[cv_i]], axis=1)
            # except:
                # pass
        
            
            
    def to_step_response(self, cv_name):
        for mv_i in self.mv:
            chip = self.matrix_dic[cv_name][mv_i]
        
            if chip.tf_card_btn.value:
                try:
                    ys_df = self.model.step_response(mv_name=mv_i, cv_name=cv_name)
                    self.temp_df = ys_df
                    chip.order, chip.gain_gui, chip.tau_gui, chip.theta_gui, chip.zeta_gui = self.model.get_model_info(mv_name=mv_i, cv_name=cv_name)
                    chip.ts, chip.tr, chip.os = self.model.get_step_info(mv_name=mv_i, cv_name=cv_name)
                    chip.update_result_dialog(df=ys_df)     
                    chip.switch_chip()
                except:
                    chip.switch_chip()
                    chip.children = [chip.no_solution_card]
            
            else:
                chip.disabled = True
                
    def to_validation(self):
        self.validation_df = DataFrame()

        for cv_i in self.model.cv:
            capsules = self.cv_label_dict[cv_i].select_validation_capsules.v_model
            validation_df = self.create_dataset(capsules)
            self.temp_df = validation_df
            yp_valid_df = self.model.predict(validation_df, cv_i)

            self.validation_df = concat([self.validation_df, yp_valid_df], axis=1)
            self.validation_df = concat([self.validation_df, validation_df[cv_i]], axis=1)  
                    

class CapsulesCard(v.Card):
    def __init__(self,
                 items=[],
                 selected_train_conditions=[],
                 selected_valid_conditions=[],
                 class_='d-flex flex-column ma-0 pa-0 pl-5 pt-5 justify-start',
                 style_='width:320px; height:200px; font-weight:bold; font-size:10pt',
                 outlined=False):
        super().__init__(class_=class_, 
                         style_=style_,
                         outlined=outlined)

        self.select_train_capsules = v.Select(tag='Training Condition',
                                        v_model=selected_train_conditions,
                                        items=items,
                                        color='success',
                                        item_color='success',
                                        dense=True,
                                        outlined=False,
                                        class_='d-flex justify-start align-start ma-0 pa-0 mt-2 mb-2',
                                        style_='width: 95%; font-size:12px',
                                        filled=True,
                                        background_color='#EDF4FF',
                                        placeholder='Select',
                                        multiple=True,
                                        hide_details=True,
                                        clearable=True,
                                        solo=True)
        
        self.select_validation_capsules = v.Select(tag='Validation Condition',
                                                   v_model=selected_valid_conditions,
                                                   items=items,
                                                   color='success',
                                                   item_color='success',
                                                   dense=True,
                                                   outlined=False,
                                                   class_='d-flex justify-start align-start ma-0 pa-0 mt-2',
                                                   style_='width: 95%; font-size:12px',
                                                   filled=True,
                                                   background_color='#EDF4FF',
                                                   placeholder='Select',
                                                   multiple=True,
                                                   hide_details=True,
                                                   clearable=True,
                                                   solo=True)
            
        self.children=['Training Condition', self.select_train_capsules,
                       'Validation Condition', self.select_validation_capsules]

        
class CVLabel(v.Btn):
    def __init__(self,
                 cv_i='',
                 items=[],
                 class_='d-flex flex-column ma-1 pa-0 justify-center align-center',
                 style_='width:170px; height:95px; font-weight:bold; font-size:10pt',
                 outlined=False,
                 color='white',
                 *args, **kwargs):
        super().__init__(class_=class_, 
                         style_=style_,
                         outlined=outlined,
                         color=color,
                         *args, **kwargs)
        self.children = [cv_i]


import ipyvuetify as v

from seeq_sysid.gui.transfer_card_setup import TransferCardSetup
from seeq_sysid.gui.transfer_card_result import TransferCardResult
from seeq_sysid.gui.utils import create_eq
from copy import deepcopy

import plotly.graph_objects as go
from plotly.express.colors import qualitative


class TransferChip(v.Card):
    tmp_cv_name = None
    card_colors = qualitative.Plotly

    n_cv = 1
    def __init__(self, 
                 mv_name=None,
                 cv_name=None,
                 elevation = 0,
                 flat=True,
                 class_='ma-1 pa-0',
                 style_='font-size:10pt;  height:96px; width:170px'):
        
        super().__init__(class_=class_,
                         elevation=elevation,
                         flat=flat,
                         style_=style_)
        
        if TransferChip.tmp_cv_name is None:
            TransferChip.tmp_cv_name = cv_name
        elif TransferChip.tmp_cv_name != cv_name:
            TransferChip.tmp_cv_name = cv_name
            TransferChip.n_cv += 1
            
        self.color = TransferChip.card_colors[TransferChip.n_cv%10-1]
        
        self.cons_eqs = v.Col(children=[], class_='d-flex flex-column ma-0 pa-0 justify-center align-center mr-1', style_='width:60px ;font-size:8pt')

        self.tf_card: TransferCardSetup = TransferCardSetup(mv_name=mv_name, cv_name=cv_name)
        self.result_dialog = TransferCardResult()

        
        fig = go.FigureWidget()
        fig['layout']['showlegend'] = False
        fig['layout']['width'] = 90
        fig['layout']['height'] = 50
        fig['layout']['xaxis']['showticklabels'] = False
        fig['layout']['yaxis']['showticklabels'] = False 
        fig['layout']['margin']['t'] = 0
        fig['layout']['margin']['b'] = 0
        fig['layout']['margin']['r'] = 0
        fig['layout']['margin']['l'] = 0
        fig['layout']['modebar']['remove'] = ['Pan', 'Zoom', 'Autoscale', 'zoomIn2d', 'zoomOut2d', 'toImage', 'resetScale2d']
        fig['layout']['yaxis2'] = {'overlaying': 'y', 'side': 'right'}
        fig['layout']['yaxis2']['showticklabels'] = False
        fig['layout']['hovermode'] = False
        
        self.blank_fig = fig
        self.fig = deepcopy(self.blank_fig)
        
        self.mv_name = self.tf_card.mv_name
        self.cv_name = self.tf_card.cv_name
              
        self.status_icon = v.Icon(children=['mdi-check-circle'], color='success', dense=True)
        self.tf_chip_items = v.Card(children=[], dense=True, color='white', class_='d-flex flex-row justify-center align-center ma-1 pa-1 mr-0', style_='height:57px; width:160px')
        self.settings_icon = v.Icon(children=['mdi-settings'], class_='mdi-18px')
        self.loading_settings = v.Icon(children=['mdi-loading'], class_='mdi-18px mdi-spin')
        self.tf_card_btn = v.Btn(children=[self.settings_icon], color='white', v_on = 'menuData.on', dense=True, class_='d-flex ma-1 pa-1', style_='height:27px; width:160px')
        self.tf_card_btn.value = True
        self.tf_card_dialog = v.Menu(value=None, children=[self.tf_card], class_='ma-2 pa-2', close_on_content_click=False, v_slots=[{
                             'name': 'activator',
                             'variable': 'menuData',
                             'children': [self.tf_card_btn]
                         }],)
        
        self.update_title()
        self.update_summary()

        # Events
        # Model equation events
        self.tf_card.order_group.on_event('change', self.update_title)
        self.tf_card.gain_checkbox.on_event('change', self.gain_checkbox_action)
        self.tf_card.deadtime_checkbox.on_event('change', self.deadtime_checkbox_action)
        self.tf_card.ramp_checkbox.on_event('change', self.ramp_checkbox_action)
        self.tf_card.signal_on_off_switch.on_event('click', self.turn_off_tf_card)
        self.result_dialog.done_btn.on_event('click', self.close_result_dialog)
        
        # Summary Events
        self.tf_card_btn.on_event('click.right', self.turn_off_tf_card)
        
        
        # Constraints events
        self.tf_card.gain_lb.on_event('change', self.update_summary)
        self.tf_card.gain_ub.on_event('change', self.update_summary)
                
        self.tf_card.tau_lb.on_event('change', self.update_summary)
        self.tf_card.tau_ub.on_event('change', self.update_summary)
        
        self.tf_card.deadtime_lb.on_event('change', self.update_summary)
        self.tf_card.deadtime_ub.on_event('change', self.update_summary)
        
        
        """ Results Chip (Fliped) """
        # Step Response Figure
        self.step_fig = None
        self.create_step_figure()
        
        # Model Info Summary
        # Gain
        self.gain_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between')
        # Time Constant (tau)
        self.tau_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between')
        # Time Delay (theta)
        self.theta_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between')    
        # Damping value (zeta)
        self.zeta_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between') 
        
        # Step Response Summary
        # Zeta Type (under damped, damped, over damped)
        self.zeta_cat = None
        self.zeta_cat_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between')
        # settling time
        self.ts = None
        self.ts_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between')
        # rise time
        self.tr = None
        self.tr_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between')
        # overshoot
        self.os = None
        self.os_text = v.Row(children=[''], style_='font-size:12pt', class_='d-flex justify-space-between')
        
        # Transfer function visualization
        self.order = None   
        self.gain_gui = None
        self.tau_gui = None
        self.theta_gui = None
        self.zeta_gui = None  
        
        self.step_info = v.Col(children=[], class_='d-flex flex-column ma-0 pa-0 justify-center align-center', style_='width:100% ;font-size:8pt', dense=True)
        self.fraction_col = v.Col(children=[], class_='d-flex flex-column ma-0 pa-0 py-1 justify-center align-center', style_='width:100% ;font-size:10pt', dense=True)
        self.model_equation = v.Card(children=[], class_='d-flex flex-row justify-center align-center ma-0 pa-0 mx-1',
                                     style_='height:30px; max-width:100%; font-size:7pt', elevation=0, flat=True, dense=True)
        
        self.info_fig = v.Card(children=[self.step_info, self.step_fig],
                               dense=True, color='white',
                               class_='d-flex flex-row justify-center align-center pa-0 ma-0 mt-1 pt-1',
                               style_='height:57px; width:160px', elevation=0, flat=True)
        
        self.results_card = v.Card(children=[self.info_fig], class_='ma-1 pa-1 pt-0 pr-0 justify-center align-center', dense=True)
        self.results_card.on_event('click', self.result_dialog_action)
        
        self.no_solution_card = v.Card(children=['No Solution Found'],
                                       dense=True, color='white', 
                                       class_='d-flex justify-center align-center pa-0 ma-0 mt-1 pt-1 ml-1 pl-1',
                                       style_='height:87px; width:160px')

        
        self.children = [self.tf_chip_items, self.tf_card_btn, self.tf_card_dialog]
        
    def turn_off_tf_card(self, *args):
        self.tf_card_btn.value = not self.tf_card_btn.value
        self.tf_card.signal_on_off_switch.v_model = not self.tf_card.signal_on_off_switch.v_model
        self.update_summary()
        
        if self.tf_card_btn.value:
            self.status_icon.color = 'success'
            self.status_icon.children=['mdi-check-circle']

            self.tf_card_btn.color = 'white'
            self.mode_select_action(self.tf_card.signal_on_off_switch)
            self.update_summary()
            self.tf_chip_items.children = [self.cons_eqs, self.fig]
            self.children = [self.tf_chip_items, self.tf_card_btn, self.tf_card_dialog]
            self.tf_chip_items.color = 'white'
            
        else:
            self.status_icon.color = 'red lighten-1'
            self.status_icon.children = ['mdi-close-circle']
            
            self.tf_card_btn.color = 'grey lighten-2'
            self.mode_select_action(self.tf_card.signal_on_off_switch)
            self.tf_chip_items.children = ['Inactive']
            self.tf_chip_items.color = 'grey lighten-2'

            
    def mode_select_action(self, item, *_):
        switch_values = ['Off', 'On']
        item.label = switch_values[item.v_model]
        
        self.tf_card.order_group.disabled = not item.v_model
        self.tf_card.ramp_checkbox.disabled = not item.v_model
        self.tf_card.gain_checkbox.disabled = not item.v_model
        self.tf_card.deadtime_checkbox.disabled = not item.v_model
        self.tf_card.gain_lb.disabled = not item.v_model
        self.tf_card.gain_ub.disabled = not item.v_model
        self.tf_card.tau_lb.disabled = not item.v_model
        self.tf_card.tau_ub.disabled = not item.v_model
        self.tf_card.deadtime_lb.disabled = not item.v_model
        self.tf_card.deadtime_ub.disabled = not item.v_model
        
        
    def update_summary(self, *args):       
        gain_lb = self.tf_card.gain_lb.v_model
        gain_ub = self.tf_card.gain_ub.v_model
        
        tau_lb = self.tf_card.tau_lb.v_model
        tau_ub = self.tf_card.tau_ub.v_model
        
        deadtime_lb = self.tf_card.deadtime_lb.v_model
        deadtime_ub = self.tf_card.deadtime_ub.v_model
        
        order = self.tf_card.order_group.v_model
        
        order_info = ''
        gain_info = ''
        tau_info = ''
        deadtime_info = ''
        children_list = []
        
        if order == 1:
            # order_eq = create_eq('FOPDT', 'black', '1.5pt', '0pt', '0pt')
            order_eq = v.Text(children=['FOPDT'], class_='ma-0 pa-0', style_='font-size:7pt; max-height:11px')
            children_list.append(order_eq)
        elif order == 2:
            order_eq = v.Text(children=['SOPDT'], class_='ma-0 pa-0', style_='font-size:7pt; max-height:11px')
            children_list.append(order_eq)
            
        if self.tf_card.gain_checkbox.v_model:
            gain_info = ' k '
            if len(gain_lb) > 0:
                gain_info = 'lb < ' + gain_info
                
            if len(gain_ub) > 0:
                gain_info = gain_info + ' < ub'
        else:
            gain_info = 'k = 1'

        gain_eq = v.Text(children=[gain_info], class_='ma-0 pa-0', style_='font-size:7pt; max-height:11px')
        children_list.append(gain_eq)
            
            
        if not self.tf_card.ramp_checkbox.v_model:
            tau_info = ' tau '
            if len(tau_lb) > 0:
                tau_info = 'lb < ' + tau_info
                
            if len(tau_ub) > 0:
                tau_info = tau_info + ' < ub'
        else:
            tau_info = 'Ramp'
        
        tau_eq = v.Text(children=[tau_info], class_='ma-0 pa-0', style_='font-size:7pt; max-height:11px')
        children_list.append(tau_eq)
            
            
        if self.tf_card.deadtime_checkbox.v_model:
            deadtime_info = ' dt '
            if len(deadtime_lb) > 0:
                deadtime_info = 'lb < ' + deadtime_info
                
            if len(deadtime_ub) > 0:
                deadtime_info = deadtime_info + ' < ub'
        else:
            deadtime_info = 'No DT'

        deadtime_eq = v.Text(children=[deadtime_info], class_='ma-0 pa-0', dense=True, style_='font-size:7pt; max-height:11px')

        children_list.append(deadtime_eq)

        
        self.cons_eqs.children = children_list
        self.tf_chip_items.children = [self.cons_eqs, self.fig]
    
    def update_title(self, *args):
        gain = self.tf_card.gain_checkbox.v_model
        deadtime = self.tf_card.deadtime_checkbox.v_model
        ramp = self.tf_card.ramp_checkbox.v_model
        order = self.tf_card.order_group.v_model
        
        k = ''
        den = ''
        theta = ''
        
        if order == 1:
            if gain:
                k = 'k_p'
            if deadtime:
                theta = 'e^{-\\theta_p s}'

            if (gain+deadtime)==0:
                k = '1'
            if ramp:
                den = 's'
            else:
                den = '\\tau_p s+1'
                
        elif order == 2:
            if gain:
                k = 'k_p'
            if deadtime:
                theta = 'e^{-\\theta_p s}'      
            if (gain+deadtime)==0:
                k = '1'
            if ramp:
                den = 's^2'
            else:
                den = '(\\tau_p s)^2+ 2 \\tau_p \\zeta s+1'
        
        tf_formula = '$\\frac{%s(s)}{%s(s)}=\\frac{%s %s}{%s}$' % (self.cv_name, self.mv_name, k, theta, den)
        self.tf_card.card_name.children = [create_eq(tf_formula, 'black', '4pt', '1pt', '1pt')]
        
        self.update_summary()
        
    def gain_checkbox_action(self, item, *args):
        self.tf_card.gain_lb.disabled = not item.v_model
        self.tf_card.gain_ub.disabled = not item.v_model
        self.update_title()
    
    def ramp_checkbox_action(self, item, *args):
        self.tf_card.tau_lb.disabled = item.v_model
        self.tf_card.tau_ub.disabled = item.v_model
        self.update_title()

    def deadtime_checkbox_action(self, item, *args):
        self.tf_card.deadtime_lb.disabled = not item.v_model
        self.tf_card.deadtime_ub.disabled = not item.v_model
        self.update_title()
        
    def close_tf_card(self, item, *args):
        self.tf_card_dialog.value = False
        
    def close_result_dialog(self, item, *args):
        self.result_dialog.v_model = None
    
    def update_meas_figure(self, df):
        self.tf_card.update_figure(df)
        self.fig = deepcopy(self.blank_fig)
        cv_plot = go.Scatter(x=df.index, y=df[self.cv_name], name=self.cv_name, yaxis='y1')
        mv_plot = go.Scatter(x=df.index, y=df[self.mv_name], name=self.mv_name, yaxis='y2', line={'dash': 'dot'})
        data = [cv_plot, mv_plot]
        self.fig.add_traces(data)
        
        self.tf_chip_items.children = [self.cons_eqs, self.fig]
        self.children = [self.tf_chip_items, self.tf_card_btn, self.tf_card_dialog]
        
    def compile_item(self):
        item = self.tf_card
        self.order = item.order_group.v_model
        self.no_gain = not item.gain_checkbox.v_model
        self.no_ramp = not item.ramp_checkbox.v_model
        self.is_dt = item.deadtime_checkbox.v_model
        
        gain_lb = item.gain_lb.v_model
        gain_ub = item.gain_ub.v_model
        
        tau_lb = item.tau_lb.v_model
        tau_ub = item.tau_ub.v_model
        
        deadtime_lb = item.deadtime_lb.v_model
        deadtime_ub = item.deadtime_ub.v_model
        
        # Gain
        self.gain_lb = float(gain_lb) if len(gain_lb) > 0 else None
        self.gain_ub = float(gain_ub) if len(gain_ub) > 0 else None
             
        # Time Constant
        self.tau_lb = float(tau_lb) if len(tau_lb) > 0 else None
        self.tau_ub = float(tau_ub) if len(tau_ub) > 0 else None
        
        # Deadtime
        self.deadtime_lb = float(deadtime_lb) if len(deadtime_lb) > 0 else None
        self.deadtime_ub = float(deadtime_ub) if len(deadtime_ub) > 0 else None
        
    
    
    """ Result Functions """
    def update_info(self): 
        # Create model prop
        self.gain_text.children = [v.Text(children=['Gain(K)'], style_='font-weight:bold'), v.Text(children=[' {:.5} '.format(self.gain_gui)] )] if self.gain_gui else []
        self.tau_text.children = [v.Text(children=['Time Constant(tau)'], style_='font-weight:bold'), v.Text(children=[' {:.5} s'.format(self.tau_gui)] )] if self.tau_gui else []
        self.theta_text.children = [v.Text(children=['Time Delay(theta)'], style_='font-weight:bold'), v.Text(children=[' {:.5} s'.format(self.theta_gui)] )] if self.theta_gui else []
        
        if self.order == 2:
            self.zeta_text.children = [v.Text(children=['Relative Damping:'], style_='font-weight:bold'), v.Text(children=[' {:.5} '.format(self.zeta_gui)])] if self.tau_gui else []
            
        self.result_dialog.model_prop_item.children = [self.gain_text, self.tau_text, self.theta_text, self.zeta_text]
              

        # Create step info
        self.ts_text.children = [v.Text(children=['Settling Time:'], style_='font-weight:bold'), v.Text(children=[' {:.5} s'.format(self.ts)] )] if self.ts else [v.Text(children=['Settling Time:'], style_='font-weight:bold'), v.Text(children=['NA'] )]
        self.tr_text.children = [v.Text(children=['Rise Time:'], style_='font-weight:bold'), v.Text(children=[' {:.5} s'.format(self.tr)] )] if self.tr else [v.Text(children=['Rise Time:'], style_='font-weight:bold'), v.Text(children=['NA'] )]
        self.os_text.children = [v.Text(children=['Overshoot:'], style_='font-weight:bold'), v.Text(children=[' {:.4} %'.format(float(self.os))] )]
        if self.order == 2:
            if (0.999 <= self.zeta_gui <= 1.001): 
                self.zeta_cat = v.Text(children=['Critically Damped'])
            elif (self.zeta_gui < 0.999):
                self.zeta_cat = v.Text(children=['Underdamped'])
            elif (self.zeta_gui > 1.001):
                self.zeta_cat = v.Text(children=['Overdamped'])
                
            self.zeta_cat_text.children = [v.Text(children=['Response Type:'], style_='font-weight:bold'), self.zeta_cat]

        self.result_dialog.step_info_item.children = [self.zeta_cat_text, self.ts_text, self.tr_text, self.os_text]
        
        # Create tf Equation
        num_str = ''
        num_str += '{0:.3}'.format(float(self.gain_gui)) if self.gain_gui else ''
        num_str += '  e^{{-{0:.3} s}}'.format(float(self.theta_gui)) if self.theta_gui else ''
        num_str = '1' if len(num_str) == 0 else num_str

        if self.order == 1:
            den_str = ''
            den_str += '{0:.5} s+1'.format(float(self.tau_gui)) if self.tau_gui else 's'

        elif self.order == 2:
            den_str = ' + {0:.5} s+1'.format(float(2*self.zeta_gui*self.tau_gui)) if (self.zeta_gui and self.tau_gui) else '+1'
            den_str = ('{0:.5} s^2'.format(float(self.tau_gui**2))+den_str) if self.tau_gui else 's^2'

        elif self.order == None:
            numerator = ''
            denominator = ''
        
        eq = '$ \\frac{%s}{%s} $' % (num_str, den_str)
        eq_widget = create_eq(eq, 'black', '4pt', '0pt', '0pt')
        self.fraction_col.children= [eq_widget]
        
        return ['G(s) = ', self.fraction_col]
    
    def update_result_dialog(self, df=None):
        self.result_dialog.model_equation.children = self.update_info()
        self.create_step_figure(df)
        
    def create_step_figure(self, df=None):
        fig: go.FigureWidget = deepcopy(self.blank_fig)
        fig['layout']['dragmode'] = False
        fig['layout']['width'] = 155
        fig['layout']['height'] = 84
        
        fig['layout']['margin']['t'] = 5
        fig['layout']['margin']['b'] = 0
        fig['layout']['margin']['r'] = 2
        fig['layout']['margin']['l'] = 2
        
        fig['layout']['xaxis']['showgrid'] = False
        fig['layout']['xaxis']['showticklabels'] = True
        fig['layout']['xaxis']['tickfont_size'] = 8
        fig['layout']['xaxis']['ticks'] = 'inside'
        
        fig['layout']['yaxis']['showgrid'] = False
        fig['layout']['yaxis']['showticklabels'] = True 
        fig['layout']['yaxis']['tickfont_size'] = 8
        fig['layout']['yaxis']['ticks'] = 'inside'

        self.step_fig = fig
        
        if df is not None:
            # Create Fig for Chip
            step_plot = go.Scatter(x=df.index, y=df[self.cv_name+'_tf'], name=self.cv_name+'_tf', yaxis='y1', line={'dash': 'solid'}, mode='lines')
            data = [step_plot]
            self.step_fig.add_traces(data)
            
            # Create Fig for Dialog
            self.result_dialog.update_figure(df[self.cv_name+'_tf'])
    
    def switch_chip(self):
        self.results_card.children = [self.step_fig]
        self.children = [self.results_card, self.result_dialog]
        
    def result_dialog_action(self, item, *args):
        self.result_dialog.v_model = True
        

    # Reset ns
    @staticmethod
    def reset():
        TransferChip.n_cv = 1



# Import Packages

import ipyvuetify as v
import ipywidgets as widgets

import plotly.express as px

from .utils import create_eq, add_tooltip


class LeftPanel(v.Card):
    colors = {
        'app_bar': '#007960',
        'controls_background': '#F6F6F6',
        'visualization_background': '#FFFFFF',
        'seeq_primary': '#007960',
    }

    additional_styles = widgets.HTML("""
        <style>
        .background_box { background-color:#007960 !important; } 
        .js-plotly-plot .plotly .modebar-btn[data-title="Produced with Plotly"] {display: none;}
        .vuetify-styles .theme--light.v-list-item .v-list-item__action-text, 
        .vuetify-styles .theme--light.v-list-item .v-list-item__subtitle {color: #212529;}
        .vuetify-styles .theme--light.v-list-item:not(.v-list-item--active):not(.v-list-item--disabled) 
        {color: #007960 !important;}
        .vuetify-styles .v-label {font-size: 14px;}
        .vuetify-styles .v-application {font-family: "Source Sans Pro","Helvetica Neue",Helvetica,Arial,sans-serif;}
        </style>""")
    v.theme.themes.light.success = '#007960'
    v.theme.themes.light.primary = '#007960'
    no_data_message = 'No plot available'

    def __init__(self,
                 model_name='Sys-Ident',
                 style_='width: 330px; min-height:94%; height:94%; max-height:98%; font-weight:bold; color:white;'
                        ' font-size:16px; border-radius:12px',
                 class_='pa-4 ma-4 ml-0 pl-7 pr-7',
                 color='#007960',
                 outlined=False,
                 shaped=False,
                 elevation=10,
                 dense=True,
                 children=None,
                 *args,
                 **kwargs):
        if children is None:
            children = []
        self.model_name = model_name

        super().__init__(tag=model_name,
                         style_=style_,
                         class_=class_,
                         color=color,
                         outlined=outlined,
                         shaped=shaped,
                         elevation=elevation,
                         dense=dense,
                         children=children,
                         **kwargs)

        # Icons
        title_icon = v.Icon(class_='', children=['mdi-tools'], color='white')

        # Drop Downs
        self.mv_select = v.Select(tag='Manipulated Variables',
                                  v_model=[],
                                  items=[],
                                  color=self.colors['seeq_primary'],
                                  item_color=self.colors['seeq_primary'],
                                  dense=True,
                                  outlined=False,
                                  class_='d-flex justify-center',
                                  style_='width: 280px; font-size:14px',
                                  filled=True,
                                  background_color='white',
                                  placeholder='Select',
                                  multiple=True,
                                  clearable=True,
                                  solo=True)

        self.cv_select = v.Select(tag='Measured Variables',
                                  v_model=[],
                                  items=[],
                                  color=self.colors['seeq_primary'],
                                  item_color=self.colors['seeq_primary'],
                                  dense=True,
                                  class_='d-flex justify-center',
                                  style_='width: 280px; font-size:14px',
                                  outlined=False,
                                  filled=True,
                                  background_color='white',
                                  placeholder='Select',
                                  multiple=True,
                                  clearable=True,
                                  solo=True)

        self.train_condition = v.Select(tag='Training Condition',
                                        v_model=[],
                                        items=[],
                                        color=self.colors['seeq_primary'],
                                        item_color=self.colors['seeq_primary'],
                                        dense=True,
                                        outlined=False,
                                        class_='d-flex justify-center',
                                        style_='width: 280px; font-size:14px',
                                        filled=True,
                                        background_color='white',
                                        placeholder='All Data',
                                        multiple=True,
                                        rounded=True,
                                        clearable=True,
                                        solo=True)

        self.validation_condition = v.Select(tag='Validation Condition',
                                             v_model=[],
                                             items=[],
                                             color=self.colors['seeq_primary'],
                                             item_color=self.colors['seeq_primary'],
                                             dense=True,
                                             class_='d-flex justify-center',
                                             style_='width: 280px; font-size:14px',
                                             outlined=False,
                                             filled=True,
                                             background_color='white',
                                             placeholder='All Data',
                                             multiple=True,
                                             rounded=True,
                                             clearable=True,
                                             solo=True)

        # Buttons
        self.identify_model_btn = v.Btn(name='identification button',
                                        style_='color:#007960; font-size:8pt; font-weight:bold; width:75px',
                                        color='white',
                                        class_='mb-4',
                                        dense=True,
                                        elevation=4,
                                        children=['Identify'],
                                        disabled=True,
                                        loading=False)

        self.validate_model_btn = v.Btn(name='validation button',
                                        style_='color:#007960; font-size:8pt; font-weight:bold; width:75px',
                                        color='white',
                                        class_='mb-4',
                                        dense=True,
                                        elevation=4,
                                        children=['Validate'],
                                        disabled=True,
                                        loading=False)

        self.push_model_btn = v.Btn(name='identification button',
                                    style_='color:#007960; font-size:8pt; font-weight:bold; width:105px',
                                    color='white',
                                    class_='mb-4',
                                    dense=True,
                                    elevation=4,
                                    children=['Push Model'],
                                    disabled=True)

        # Card
        self.identify_push_card = v.Card(
            children=[self.identify_model_btn, self.validate_model_btn, self.push_model_btn],
            style_='background:none', flat=True, class_='d-flex justify-space-between')
        
        self.export_model_btn = v.Btn(children=[v.Icon(children=['mdi-export-variant'], color='white')], icon=True)
        self.export_model_btn_tp = add_tooltip(self.export_model_btn, 'Export Model')

        # Title
        self.title = v.Card(class_='pt-5 mb-7 mx-0 d-flex justify-left align-center', style_='font-size:20px; background:none',
                            dark=True, center=True, align='center', flat=True,
                            children=[title_icon, v.Divider(vertical=True, class_='mx-2'),
                                      "{} Settings".format(model_name),
                                      v.Spacer(),
                                      self.export_model_btn])

        # Set Tooltips

        # Set Panel Children
        self.children = [self.title,
                         'Manipulated Variables (MV)', self.mv_select,
                         'Measured Variables (CV)', self.cv_select,
                         #                          v.Divider(class_='mb-4'),
                         'Training Conditions', self.train_condition,
                         'Validation Conditions', self.validation_condition,
                         v.Divider(class_='mb-4'),
                         self.identify_push_card]


# ARX Model Panel
class ARXPanel(LeftPanel):
    def __init__(self, *args, **kwargs):
        model_name = 'TimeSeries'
        super().__init__(model_name=model_name,
                         *args, **kwargs)

        # Icons
        title_icon = v.Icon(class_='', children=['mdi-tools'], color='white')

        # Expanding Panel
        # ARX ModelStructure
        self.model_struct_select = v.Select(v_model='ARX',
                                            items=['ARX', 'FIR', 'ARIMAX'],
                                            color=self.colors['seeq_primary'],
                                            item_color=self.colors['seeq_primary'],
                                            dense=True,
                                            outlined=False,
                                            class_='pl-3 my-0 py-0',
                                            style_='width: 150px; font-size:14px',
                                            filled=True,
                                            align='top',
                                            background_color='white',
                                            placeholder='Select',
                                            multiple=False,
                                            clearable=False,
                                            solo=True)
        self.model_struct_select.on_event('change', self.model_struct_action)

        self.model_struct = v.Row(children=[v.Row(children=['Type: ', v.Spacer()],
                                                  dense=True,
                                                  align='top',
                                                  class_='mt-2',
                                                  no_gutters=True,
                                                  style_='font-weight:bold; color:white; font-size:13px'),
                                            self.model_struct_select],
                                  class_='my-0 py-0',
                                  dense=True,
                                  style_='font-weight:bold; color:white; font-size:13px',
                                  align='top',
                                  no_gutters=True
                                  )

        self.na_min = v.TextField(label='min', v_model='2', dense=True, class_='pl-2', color='white', dark=True,
                                  style_='width:30px')
        self.na_max = v.TextField(label='max', v_model='2', dense=True, class_='pl-2', color='white', dark=True,
                                  style_='width:30px')
        self.na = v.Row(
            children=[v.Row(children=['Auto-Regressive', create_eq('$(n_a):$', 'white', 2, top='0px')], class_='mt-0',
                            no_gutters=True, ),
                      self.na_min, self.na_max],
            no_gutters=True,
            class_='',
            dense=True,
            style_='font-weight:bold; color:white; font-size:13px', align='center')

        self.nb_min = v.TextField(label='min', v_model='2', dense=True, class_='pl-2', color='white', dark=True,
                                  style_='width:30px', align='top')
        self.nb_max = v.TextField(label='max', v_model='2', dense=True, class_='pl-2', color='white', dark=True,
                                  style_='width:30px', align='top')
        self.nb = v.Row(no_gutters=True,
                        children=[v.Row(children=['Exogenous Input', create_eq('$(n_b):$', 'white', 2, top='0px')],
                                        class_='mt-0', no_gutters=True, ),
                                  self.nb_min, self.nb_max],
                        class_='',
                        dense=True,
                        style_='font-weight:bold; color:white; font-size:13px',
                        align='center')

        self.nk_min = v.TextField(label='min', v_model='0', dense=True, class_='pl-2', color='white', dark=True,
                                  style_='width:15px', align='top')
        self.nk_max = v.TextField(label='max', v_model='0', dense=True, class_='pl-2 mr-1', color='white', dark=True,
                                  style_='width:15px', align='top')
        self.nk = v.Row(children=[v.Row(children=['Input Delay', create_eq('$(n_k):$', 'white', 2, top='0px')],
                                        class_='mt-0', no_gutters=True, ), v.Spacer(), self.nk_min, self.nk_max],
                        class_='d-flex justify-right',
                        dense=True,
                        style_='font-weight:bold; color:white; font-size:13px', align='center')

        self.orders_panel_obj = v.ExpansionPanel(children=[
            v.ExpansionPanelHeader(children=['Model Structure'],
                                   style_='font-weight:bold; color:white; font-size:12pt',
                                   class_='my-0 py-0',
                                   dense=True,
                                   no_gutters=True,
                                   align='center',
                                   dark=True),
            v.ExpansionPanelContent(
                children=[v.Col(children=[
                    self.model_struct,
                    self.na, self.nb, self.nk],
                    style_='font-size:14px; font-weight:bold',
                    dark=True,
                    align='center',
                    no_gutters=True,
                    class_='my-0 py-0 px-0 ml-0',
                    color='white',
                    dense=True)])],
            style_='background-color:#007960')

        self.orders_panel = v.ExpansionPanels(children=[self.orders_panel_obj], dense=True, style_='width: 300px',
                                              flat=False)
        self.orders_layout = v.Layout(children=[self.orders_panel], class_='mb-6', dense=True, flat=False)

        # Arx
        self.title = v.Card(class_='pt-5 mb-7 mx-0 d-flex justify-left align-center', style_='font-size:20px; background:none',
                            dark=True, center=True, align='center', flat=True,
                            children=[title_icon, v.Divider(vertical=True, class_='mx-2'),
                                      "{} Settings".format('Time Series'),
                                      v.Spacer(),
                                      self.export_model_btn_tp
                                     ])

        self.children = [self.title,
                         'Manipulated Variables (MV)', self.mv_select,
                         'Measured Variables (CV)', self.cv_select,
                         self.orders_layout,
                         #                          v.Divider(class_='mb-4'),
                         'Training Conditions', self.train_condition,
                         'Validation Conditions', self.validation_condition,
                         #                          v.Divider(class_='mb-6'),
                         self.identify_push_card]

    def model_struct_action(self, item, *_):
        if item.v_model == 'ARX':
            self.na_min.v_model = '2'
            self.na_max.v_model = '2'
            self.na_min.disabled = False
            self.na_max.disabled = False

        if item.v_model == 'FIR':
            self.na_min.v_model = '0'
            self.na_max.v_model = '0'
            self.na_min.disabled = True
            self.na_max.disabled = True


# State-Space Panel
class SSPanel(LeftPanel):
    def __init__(self, *args, **kwargs):
        model_name = 'Subspace'
        super().__init__(model_name=model_name,
                         *args, **kwargs)

        # Icons
        title_icon = v.Icon(class_='', children=['mdi-tools'], color='white')

        # Drop Downs
        self.method_select = v.Select(tag='Methods',
                                      v_model='Least Square',
                                      # items=['N4SID', 'DMDc'],
                                      items=['Least Square'],
                                      color=self.colors['seeq_primary'],
                                      item_color=self.colors['seeq_primary'],
                                      dense=True,
                                      outlined=False,
                                      class_='d-flex justify-center',
                                      style_='width: 280px; font-size:14px',
                                      filled=True,
                                      background_color='white',
                                      placeholder='Select',
                                      multiple=False,
                                      solo=True)

        self.method_select.on_event('change', self.method_select_action)

        # Expanding Panel
        # State-Space Model Residual Energy Threshold
        self.method_box = None
        self.threshold_box = v.TextField(label='epsilon', v_model='1e-6', dense=True, class_='pl-2 pt-1', color='white',
                                         dark=True, style_='width:70px; font-size:11pt', hint='ex: 1e-6')
        self.threshold = v.Row(
            children=[v.Row(children=['Threshold'], class_='mt-0', style_='font-size:11pt; font-weight:bold',
                            no_gutters=True),
                      self.threshold_box],
            class_='',
            dense=True,
            no_gutters=True,
            style_='font-weight:bold; color:white; font-size:13px', align='center')

        self.order_box = v.TextField(label='Order', v_model='4', dense=True, class_='pl-2 pt-1', color='white',
                                     dark=True, style_='width:70px; font-size:11pt', hint='ex: 4')
        self.order = v.Row(children=[
            v.Row(children=['States', create_eq('$(n):$', 'white', 2, top='0px')],
                  class_='mt-0',
                  no_gutters=True,
                  style_='font-size:11pt; font-weight:bold'), self.order_box],
            class_='',
            dense=True,
            style_='font-weight:bold; color:white; font-size:13px',
            align='center',
            no_gutters=True)

        self.multiplier_min = v.TextField(label='min', v_model='1', dense=True, class_='pl-2', color='white', dark=True,
                                          style_='width:30px', align='top', hint='>=1')
        self.multiplier_max = v.TextField(label='max', v_model='5', dense=True, class_='pl-2', color='white', dark=True,
                                          style_='width:30px', align='top')
        self.multiplier = v.Row(no_gutters=True,
                                children=[v.Row(children=['Order Multiplier'], class_='mt-0', no_gutters=True, ),
                                          self.multiplier_min, self.multiplier_max],
                                class_='my-0 py-0',
                                dense=True,
                                style_='font-weight:bold; color:white; font-size:13px',
                                align='center')

        self.method_box = self.multiplier
        
        self.shift_type = v.Select(tag='Measured Variables',
                                  v_model=[],
                                  items=[],
                                  color=self.colors['seeq_primary'],
                                  item_color=self.colors['seeq_primary'],
                                  dense=True,
                                  class_='ml-4 pl-4 pt-4 my-0 py-0',
                                  style_='max-width:150px; font-size:10pt',
                                  outlined=False,
                                  filled=True,
                                  background_color='white',
                                  placeholder='Select',
                                  multiple=False,
                                  solo=True)
        self.shift_type.items = ['Initial', 'Mean']
        self.shift_type.v_model= self.shift_type.items[0]
        self.shift_type_label = v.Text(children=['Shift'], class_='white--text', style_='font-weight:bold', dense=True)

        self.shift_type_layout = v.Layout(children=[self.shift_type_label, self.shift_type], 
                                     dense=True,
                                     style_='width:280px; height:45px; overflow: hidden',
                                     class_='d-flex flex-row justify-start align-center pt-3')

        self.orders_panel_obj = v.ExpansionPanel(
            children=[v.ExpansionPanelHeader(children=['Model Order'],
                                             style_='font-weight:bold; color:white; font-size:12pt',
                                             dense=True,
                                             align='center',
                                             dark=True),
                      v.ExpansionPanelContent(
                          children=[v.Col(children=[self.method_box, self.shift_type_layout],
                                          style_='font-size:14px; font-weight:bold',
                                          dark=True,
                                          align='center',
                                          dense=True)])],
            style_='background-color:#007960')

        self.orders_panel = v.ExpansionPanels(children=[self.orders_panel_obj], dense=True, style_='width: 300px',
                                              flat=False)
        self.orders_layout = v.Layout(children=[self.orders_panel], class_='mb-6', dense=True, flat=False)

        # Sub-Space
        self.title = v.Card(class_='pt-5 mb-7 mx-0 d-flex justify-left align-center', style_='font-size:20px; background:none',
                            dark=True, center=True, align='center', flat=True,
                            children=[title_icon, v.Divider(vertical=True, class_='mx-2'),
                                      '{} Settings'.format(self.model_name)])

        self.children = [self.title,
                         'Manipulated Variables (MV)', self.mv_select,
                         'Measured Variables (CV)', self.cv_select,
                         # 'Method', self.method_select,
                         self.orders_layout,
                         #                          v.Divider(class_='mb-4'),
                         'Training Conditions', self.train_condition,
                         'Validation Conditions', self.validation_condition,
                         v.Divider(class_='mb-6'),
                         self.identify_push_card]

    def method_select_action(self, item, *_):
        if item.v_model == 'N4SID':
            self.method_box = self.threshold
            self.orders_panel_obj = v.ExpansionPanel(children=[
                v.ExpansionPanelHeader(children=['Model Order'],
                                       style_='font-weight:bold; color:white; font-size:12pt',
                                       dense=True,
                                       align='center',
                                       dark=True),
                v.ExpansionPanelContent(
                    children=[v.Col(children=[self.method_box],
                                    style_='font-size:14px; font-weight:bold',
                                    dark=True,
                                    align='center',
                                    dense=True)])],
                style_='background-color:#007960')

            self.orders_panel = v.ExpansionPanels(children=[self.orders_panel_obj], dense=True, style_='width: 300px',
                                                  flat=False)
            self.orders_layout = v.Layout(children=[self.orders_panel], class_='mb-6', dense=True, flat=False)
            self.children = [self.title,
                             'Manipulated Variables (MV)', self.mv_select,
                             'Measured Variables (CV)', self.cv_select,
                             # 'Method', self.method_select,
                             self.orders_layout,
                             v.Divider(class_='mb-4'),
                             'Training Conditions', self.train_condition,
                             'Validation Conditions', self.validation_condition,
                             v.Divider(class_='mb-6'),
                             self.identify_push_card]

        elif item.v_model == 'DMDc':
            self.method_box = self.order
            self.orders_panel_obj = v.ExpansionPanel(children=[
                v.ExpansionPanelHeader(children=['Model Order'],
                                       style_='font-weight:bold; color:white; font-size:12pt',
                                       dense=True,
                                       align='center',
                                       dark=True),
                v.ExpansionPanelContent(
                    children=[v.Col(children=[self.method_box],
                                    style_='font-size:14px; font-weight:bold',
                                    dark=True,
                                    align='center',
                                    dense=True)])],
                style_='background-color:#007960')

            self.orders_panel = v.ExpansionPanels(children=[self.orders_panel_obj], dense=True, style_='width: 300px',
                                                  flat=False)
            self.orders_layout = v.Layout(children=[self.orders_panel], class_='mb-6', dense=True, flat=False)
            self.children = [self.title,
                             'Manipulated Variables (MV)', self.mv_select,
                             'Measured Variables (CV)', self.cv_select,
                             # 'Method', self.method_select,
                             self.orders_layout,
                             #                              v.Divider(class_='mb-4'),
                             'Training Conditions', self.train_condition,
                             'Validation Conditions', self.validation_condition,
                             v.Divider(class_='mb-6'),
                             self.identify_push_card]


# Neural Network Panel
class NNPanel(LeftPanel):
    def __init__(self, *args, **kwargs):
        model_name = 'NeuralNetwork'
        super().__init__(model_name=model_name,
                         *args, **kwargs)

        # Icons
        title_icon = v.Icon(class_='', children=['mdi-tools'], color='white')

        # Mode Switch
        switch_values = ['Manual', 'Auto']
        self.options_switch = v.Switch(tag='switch',
                                       v_model=True,
                                       inset=True,
                                       label='Manual',
                                       dark=False,
                                       vertical=True,
                                       dense=True,
                                       )

        self.options_switch.label = switch_values[self.options_switch.v_model]
        self.options_switch.on_event('change', self.mode_select_action)
        self.mode_row = v.Card(children=['Mode', v.Spacer(), self.options_switch],
                               color='none',
                               flat=True,
                               class_='d-flex justify-center flex-row align-center pt-1 mx-1',
                               style_='font-size:12pt; font-weight:bold; height:40px',
                               dense=True)

        # Auto Mode Widgets
        help_table_items = [{'Feature\\Mode': 'Optimal number of units for hidden layer(s)',
                             'Low': 'Yes',
                             'Normal': 'Yes',
                             'High': 'Yes'},
                            {'Feature\\Mode': 'Optimal number of hidden layers',
                             'Low': 'No',
                             'Normal': 'Yes',
                             'High': 'Yes'},
                            {'Feature\\Mode': 'Optimal batch size',
                             'Low': 'No',
                             'Normal': 'No',
                             'High': 'Yes'}]
        columns = [
            {'text': 'Feature\Mode', 'sortable': False, 'value': 'Feature\Mode'},
            {'text': 'Low', 'sortable': False, 'value': 'Low', 'align': 'center'},
            {'text': 'Normal', 'sortable': False, 'value': 'Normal', 'align': 'center'},
            {'text': 'High', 'sortable': False, 'value': 'High', 'align': 'center'},
        ]
        self.help_table = v.DataTable()
        self.help_table.headers = columns
        self.help_table.items = help_table_items
        self.help_table.hide_default_footer = True
        self.help_table.disable_sort = True

        self.slider_help_tip_btn = v.Btn(icon=True, children=[v.Icon(children=['mdi-help-circle-outline'])])
        self.slider_help_tip_btn.on_event('click', self.help_action)

        self.close_help_dialog_btn = v.Btn(children=['CLOSE'], color='#007960', text=True)
        self.close_help_dialog_btn.on_event('click', self.close_help_action)

        help_card_content = [
            self.help_table,
            self.close_help_dialog_btn
        ]

        help_dialog_card = v.Card(children=help_card_content, class_='d-flex flex-column justify-right pa-2 ma-3 my-0',
                                  flat=True)

        self.help_dialog = v.Dialog(name='OpneWB',
                                    children=[v.Card(children=[help_dialog_card])],
                                    v_model=False,
                                    max_width='600px')
        self.help_dialog.on_event('keydown.stop', lambda *args: None)

        self.auto_slider_title = v.Row(children=[v.Icon(children=['mdi-brain'], class_='px-2 pb-0 mb-0'),
                                                 'Computation Cost',
                                                 v.Spacer(),
                                                 self.help_dialog,
                                                 self.slider_help_tip_btn],
                                       align='center',
                                       class_='pb-0 pt-1 px-1',
                                       style_='font-size:11pt')
        self.auto_mode_slider = v.Slider(tick_labels=['Low', 'Medium', 'High'],
                                         max=2,
                                         v_model='0',
                                         style_='font-size:10pt; font-weight:bold',
                                         class_='',
                                         dense=True,
                                         dark=False)
        self.auto_mode_list = [self.auto_slider_title, self.auto_mode_slider]

        # Manual Mode Widgets (Coming Soon...)
        self.manual_btn_title = v.Row(
            children=[v.Icon(children=['mdi-graph-outline'], class_='px-2 pb-0 mb-1'), 'Custom Neural Network'],
            class_='pb-0 pt-1 px-1',
            style_='font-size:11pt')
        self.manual_mode_btn = v.Btn(children=['Coming Soon...'], class_='d-flex justify-center my-1', align='center',
                                     dark=False, disabled=True)
        self.manual_mode_btn.on_event('click', self.manual_mode_btn_action)
        self.manual_mode_list = [self.manual_btn_title, self.manual_mode_btn]

        self.custom_nn_card = v.Card(color='white', children=['Hello World'], height='80%', align='center',
                                     class_='d-flex flex-row justify-center')
        self.custom_nn_dialog = v.Dialog(v_model=False,
                                         align='center',
                                         width='50%',
                                         class_='d-flex flex-row justify-center align-center',
                                         align_centered=True,
                                         children=[self.custom_nn_card])

        self.custom_nn_dialog.on_event('keydown.stop', lambda *args: None)

        # Mode Card
        self.mode_card = v.Card(flat=True, class_='d-flex flex-column justify-center py-1', height='80px')
        self.mode_card.children = self.auto_mode_list

        self.switch_card = v.Card(children=[self.mode_row,
                                            v.Divider(class_='my-2'),
                                            self.mode_card],
                                  color='white', class_='px-3 py-1 mb-4 mt-0')

        # Neural Network
        self.title = v.Card(class_='pt-5 mb-7 mx-0 d-flex justify-left align-center', style_='font-size:20px; background:none',
                            dark=True, center=True, align='center', flat=True,
                            children=[title_icon, v.Divider(vertical=True, class_='mx-2'),
                                      '{} Settings'.format('Neural Network')])

        self.children = [self.title,
                         'Manipulated Variables (MV)', self.mv_select,
                         'Measured Variables (CV)', self.cv_select,
                         v.Divider(class_='mb-4'),
                         self.switch_card,
                         self.custom_nn_dialog,
                         v.Divider(class_='mb-4'),
                         'Training Conditions', self.train_condition,
                         'Validation Conditions', self.validation_condition,
                         v.Divider(class_='mb-6'),
                         self.identify_push_card]

    def mode_select_action(self, item, *_):
        switch_values = ['Manual', 'Auto']
        item.v_model != item.v_model
        item.label = switch_values[item.v_model]
        if item.label == 'Manual':
            self.mode_card.children = self.manual_mode_list

        elif item.label == 'Auto':
            self.mode_card.children = self.auto_mode_list

    def manual_mode_btn_action(self, *_):
        self.custom_nn_dialog.v_model = True

    def help_action(self, *args):
        self.help_dialog.v_model = True

    def close_help_action(self, *args):
        self.help_dialog.v_model = None



from os import environ
environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from pandas import DataFrame, concat
from numpy import array
from tensorflow import get_logger
get_logger().setLevel('ERROR')

from tensorflow.keras import layers, callbacks, Sequential, optimizers
from keras_tuner import RandomSearch, BayesianOptimization, HyperModel, Hyperband

from .base import Model


# Neural Network (RNN)
class NN(Model):
    def __init__(self):
        super().__init__()

        self.p = None
        self.q = None
        self.nn_type = 'RNN'
        self.mode = None
        self.options = None
        self.tuner = None
        self.units = None
        self.n_layers = None
        self.batch_size = None
        self.max_trials = None

        self.model = None
        self.history = None
        self.window = 10
        self.Min = None
        self.Max = None

    def normalize(self, df: DataFrame = None):
        df_norm = (df - self.Min) / (self.Max - self.Min)
        return df_norm

    def denormalize(self, df_norm: DataFrame = None):
        df = df_norm[self.cv] * (self.Max[self.cv] - self.Min[self.cv]) + self.Min[self.cv]
        return df

    def create_snapshot(self, in_df: DataFrame = None, out_df: DataFrame = None):
        n_step = len(in_df)

        x_train = []
        y_train = []

        if in_df is not None:
            for i in range(self.window, n_step):
                x_train.append(in_df.iloc[i - self.window:i].to_numpy())
            x_train = array(x_train)

        if out_df is not None:
            for i in range(self.window, n_step):
                y_train.append(out_df.iloc[i].to_numpy())
            y_train = array(y_train)

        return x_train, y_train

    def option_maker(self):
        if self.mode == 0:  # Mode = Low
            self.units = list(range(5, 206, 20))
            self.n_layers = [1]
            self.batch_size = [32]
            self.max_trials = 6

        elif self.mode == 1:  # Mode = Medium
            self.units = list(range(5, 206, 20))
            self.n_layers = [1, 2, 3]
            self.batch_size = [32]
            self.max_trials = 15

        elif self.mode == 2:  # Mode = High
            self.units = list(range(5, 206, 20))
            self.n_layers = [1, 2, 3]
            self.batch_size = [32, 64, 128]
            self.max_trials = 30

    def identify(self, df: DataFrame = None):
        # Set Mode Options
        self.option_maker()

        # Early Stopping
        tune_es = callbacks.EarlyStopping(monitor='val_loss', patience=3)
        fit_es = callbacks.EarlyStopping(monitor='val_loss', patience=15)

        # Number of Inputs & Outputs
        self.p = len(self.mv + self.cv)
        self.q = len(self.cv)

        # Normalize data
        df_norm = self.normalize(df=df)

        # Input/Output Data
        in_df = df_norm[self.mv + self.cv]
        out_df = df_norm[self.cv]

        # Data Label
        self.label = [tag_label + '_nn' for tag_label in out_df.columns]

        # Create Data Snapshots for NN
        x_train, y_train = self.create_snapshot(in_df=in_df, out_df=out_df)

        # Setup Hyperparameter Optimizer
        hypermodel = HyperNN(units=self.units, n_layers=self.n_layers, p=self.p, q=self.q, window=self.window)
        tuner = HyperTuner(
            auto_bs=self.batch_size,
            hypermodel=hypermodel,
            objective="val_loss",
            max_trials=self.max_trials,
            num_initial_points=2,
            alpha=1e-4,
            beta=2.6,
            overwrite=True,
            directory="temp_dir",
            project_name="Seeq_NN",
        )

        tuner.oracle.multi_worker = True

        # Run Hyperparameter Optimizer
        tuner.search(x_train, y_train,
                     epochs=20, validation_split=0.2,
                     verbose=0, callbacks=[tune_es])
        self.tuner = tuner

        # Get the optimal hyperparameters
        best_hps = tuner.get_best_hyperparameters(num_trials=1)[0]

        model = tuner.hypermodel.build(best_hps)

        # Training Performance
        self.history = model.fit(x_train, y_train, epochs=1000, validation_split=0.2, callbacks=[fit_es], verbose=0)

        # Save Model
        self.model = model

        self.status = True

    def forecast(self, df: DataFrame = None):
        df_norm = self.normalize(df=df)

        in_df = df_norm[self.mv+self.cv]
        x_train, _ = self.create_snapshot(in_df=in_df)

        yp_norm = self.model.predict(x_train)

        yp_norm_df = DataFrame(yp_norm, columns=self.cv)
        yp_df = self.denormalize(yp_norm_df)
        dummy_rows = df[self.cv].iloc[:self.window].shift(self.window)
        yp_df = concat([dummy_rows, yp_df], ignore_index=True)
        yp_df.fillna(method='bfill', inplace=True)
        yp_df.columns = self.label

        return yp_df


class HyperNN(HyperModel):
    def __init__(self, units: list, n_layers: list, p: int, q: int, window: int, *args, **kwargs):
        self.units = units
        self.n_layers = n_layers
        self.p = p
        self.q = q
        self.window = window
        super(HyperNN, self).__init__(*args, **kwargs)

    def build(self, hp):
        model = Sequential()
        n_layers = hp.Choice(name='n_layers', values=self.n_layers)
        for i_layer in range(n_layers):
            layer_name = "units{}".format(i_layer)
            model.add(
                layers.SimpleRNN(
                    input_shape=(self.window, self.p),
                    units=hp.Choice(layer_name, values=self.units),
                    return_sequences=True,
                )
            )
            model.add(layers.Dropout(0.25))
        model.add(layers.Flatten())
        model.add(layers.Dense(self.q,
                               activation="linear"))
        model.compile(
            optimizer=optimizers.Adam(),
            loss="MSE",
            metrics=["accuracy"]
        )

        return model


class HyperTuner(BayesianOptimization):
    def __init__(self, auto_bs: list, *args, **kwargs):
        self.auto_bs = auto_bs
        super(HyperTuner, self).__init__(*args, **kwargs)

    def run_trial(self, trial, *args, **kwargs):
        kwargs['batch_size'] = trial.hyperparameters.Choice('batch_size', self.auto_bs)
        super(HyperTuner, self).run_trial(trial, *args, **kwargs)



from gekko import GEKKO
from seeq import spy

from numpy import vstack, zeros, reshape, array, zeros_like, diff
from pandas import DataFrame

from .base import Model
from re import split
from scipy.linalg import lstsq


def create_formula_variable_name(names):
    formula_name = []
    for item in names:
        item = item.lower()
        item = split("_| | ", item)
        f_name = '$'
        
        cnt = 0
        for i in item:
            if cnt > 0:
                f_name += i[0]
            else:
                f_name += i
            cnt += 1
        formula_name.append(f_name)

    return formula_name


class ARX(Model):
    """ AutoRegressive with Exogenous input (ARX)
    Model Structure:
    y(k+1) = a_0*y(k)+a_1*y(k-1)+...+a_(na-1)*y(k-na+1)
              + b_0+u(k)+b_1*u(k-1)+...+b_(nb-1)*u(k-na-nk+1)
    where:
        u: Input(s)   [pandas DataFrame]
        y: Output(s)  [pandas DataFrame]
        na: AutoRegressive term order
            na=2 -> a_0*y(k) + a_1*y(k-1)
        nb: Exogenous input order
            nb=2 -> b_0*u(k) + b_1*u(k-1)
        nk: Input Delay
            nk=2 -> b_0*u(k-2) + b_1*u(k-3)
        n_u: Number of Inputs
        n_y: Number of Outputs
    """

    def __init__(self):

        super().__init__()
        self.model_struct = 'ARX'

        self.best_model = None
        self.p = None
        # AR part order
        self.na_min = 0
        self.na_max = 2
        self.na = None
        # Exogenous part order
        self.nb_min = 0
        self.nb_max = 2
        self.nb = None
        # Input delay
        self.nk_min = 0
        self.nk_max = 0
        self.nk = None
        
        self.e_AR = None
        self.nh = None
        self.nh_min = 0
        self.nh_max = 1
        self.nv = None
        self.nv_min = 1
        self.nv_max = 2

    def identify(self, df: DataFrame = None):
        self.error_best = 1e9

        self.df = df
        u_df = df[self.mv]
        y_df = df[self.cv]
        ly, ny = y_df.shape
        lu, nu = u_df.shape

        t = list(range(len(y_df.index)))
        
        if 'MA' in self.model_struct:
            shift='mean'
        else:
            shift = 'calc'

        for i_a in range(self.na_min, self.na_max + 1):
            for i_b in range(self.nb_min, self.nb_max + 1):
                for i_k in range(self.nk_min, self.nk_max + 1):
                    model = GEKKO(remote=False)\
                    
                    # AR Model
                    yp, p, K = model.sysid(t=t, u=u_df, y=y_df, na=i_a, nb=i_b, nk=i_k, pred='meas', shift=shift)

                    e = y_df.to_numpy() - yp
                    E = sum(sum(e**2)) / len(e**2)

                    if E < self.error_best:
                        self.best_model = model
                        self.error_best = E
                        self.e_AR = e
                        self.p = p
                        self.na = i_a
                        self.nb = i_b
                        self.nk = i_k
                        self.yp = yp
                        self.K = K
                        
        # ARIMAX
        if (self.model_struct == 'ARIMAX') or (self.model_struct == 'ARMAX'):
            self.res_best = 1e9

            for nv in range(self.nv_min, self.nv_max+1):
                for nh in range(self.nh_min, self.nh_max+1):
                    ma_coe = [0]*ny
                    
                    e = diff(self.e_AR, nh, axis=0)

                    for y_i in range(ny):

                        x = []
                        y = []

                        x_i_df = DataFrame()
                        y_i_df = DataFrame()

                        for i in range(nv):
                            x_i_df['e_d'+str(i+1)] = e[nv-i-1:-1-i, y_i]
                        y_i_df['e'] = e[nv:, y_i]

                        x = x_i_df.to_numpy()
                        y = y_i_df.to_numpy()
                        
                        self.temp_x = x
                        self.temp_y = y

                        ma_coe[y_i], res = lstsq(x, y, cond=None)[0:2]

                        if sum(res) <= self.res_best:
                            self.nv = nv
                            self.nh = nh
                            self.p['d'] = ma_coe
                    
            e = diff(self.e_AR, self.nh, axis=0)
            
            x = []
            y_ma = [0]*ny
            
            for y_i in range(ny):
            
                x_i_df = DataFrame()
                for nv_i in range(self.nv):
                    x_i_df['e_d'+str(nv_i+1)] = e[self.nv-nv_i-1:-nv_i-1, y_i]

                x = x_i_df.to_numpy()

                y_ma[y_i] = x.dot(self.p['d'][y_i].flatten())

                self.yp[self.nv+self.nh:, y_i] = array(y_ma[y_i]).T + self.yp[self.nv+self.nh:, y_i]


        self.status = True

    def predict(self, df: DataFrame = None):
        u_df = df[self.mv]
        y_df = df[self.cv]

        ly, ny = y_df.shape
        lu, nu = u_df.shape
        na = self.na
        nb = self.nb
        nk = self.nk

        alpha = reshape(self.p['a'].transpose(), [ny, na])
        beta = self.p['b']
        c = self.p['c']
        if 'MA' in self.model_struct:
            eps = self.p['d']

        nd_max = max(na, nb + nk)

        yp: array = y_df.iloc[:nd_max].to_numpy()
        ni = max(0, na - nb - nk)

        if self.model_struct == 'ARX':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    y_p[i] = alpha[i][::-1].dot(y_df[k:k + na].iloc[:, [i]]) + c[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])

                yp = vstack([yp, y_p])

        elif self.model_struct == 'FIR':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    # y_p[i] = C[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])

                yp = vstack([yp, y_p])
                
        if (self.model_struct == 'ARIMAX') or (self.model_struct == 'ARMAX'):
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    y_p[i] = alpha[i][::-1].dot(y_df[k:k + na].iloc[:, [i]]) + c[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])

                yp = vstack([yp, y_p])
                
            e = y_df.to_numpy() - yp
             
            e = diff(e, self.nh, axis=0)

            x = []
            y_ma = [0]*ny
            
            for y_i in range(ny):
            
                x_i_df = DataFrame()
                for nv_i in range(self.nv):
                    x_i_df['e_d'+str(nv_i+1)] = e[self.nv-nv_i-1:-nv_i-1, y_i]

                x = x_i_df.to_numpy()

                y_ma[y_i] = x.dot(self.p['d'][y_i].flatten())
            
                yp[self.nv+self.nh:, y_i] = array(y_ma[y_i]).T + yp[self.nv+self.nh:, y_i]

        labels = []
        for label in y_df.columns:
            labels.append(label + '_pred')

        yp = DataFrame(yp, columns=labels)
        yp.fillna(method='ffill', inplace=True)
        yp.fillna(method='bfill', inplace=True)
        self.yp = yp

        return yp

    def forecast(self, df: DataFrame = None):
        if 'MA' in self.model_struct:
            yp = self.predict(df)
            return yp
    
        u_df = df[self.mv]
        y_df = df[self.cv]

        ly, ny = y_df.shape
        lu, nu = u_df.shape
        na = self.na
        nb = self.nb
        nk = self.nk

        alpha = reshape(self.p['a'].transpose(), [ny, na])
        beta = self.p['b']
        c = self.p['c']

        nd_max = max(na, nb + nk)

        yp = y_df.iloc[:nd_max].to_numpy()
        yo = yp[-na:].transpose()

        ni = max(0, na - nb - nk)

        if self.model_struct == 'ARX':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    y_p[i] = alpha[i][::-1].dot(yo[i]) + c[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])
                    yo[i][:na - 1] = yo[i][1:]
                    yo[i][-1] = y_p[i]
                yp = vstack([yp, y_p])

        elif self.model_struct == 'FIR':
            for k in range(ni, lu - nb - nk):
                y_p = zeros(ny)
                for i in range(ny):
                    y_p[i] = c[i]
                    for j in range(nu):
                        y_p[i] += beta[i].transpose()[j][::-1].dot(u_df[k:k + nb + nk].iloc[:, [j]])
                    yo[i][:na - 1] = yo[i][1:]
                    yo[i][-1] = y_p[i]
                yp = vstack([yp, y_p])

        self.label = [tag_label + '_pred' for tag_label in y_df.columns]

        yp_df = DataFrame(yp, columns=self.label)
        yp_df.fillna(method='ffill', inplace=True)
        yp_df.fillna(method='bfill', inplace=True)
        return yp_df

    def create_formula(self, tags, signal_df, **kwargs):
        u_name = self.mv
        y_name = self.cv

        n_y = self.p['a'].shape[1]
        n_u = self.p['b'].shape[2]
        n_a = self.p['a'].shape[0]
        n_bk = self.p['b'].shape[1]

        yf_name = create_formula_variable_name(y_name)
        uf_name = create_formula_variable_name(u_name)

        formula_dic = {}
        for tag in range(len(yf_name)):
            formula_dic[yf_name[tag]] = tags[tags['Name'] == y_name[tag]]

        for tag in range(len(uf_name)):
            formula_dic[uf_name[tag]] = tags[tags['Name'] == u_name[tag]]

        time_interval = self.df.index[1] - self.df.index[0]

        timestep: float = time_interval.total_seconds()

        formula_list = []

        for i in range(n_y):
            formula = ''
            for j in range(n_a):
                formula += ' {}.move({}s)*({})\n+'.format(yf_name[i], (j + 1) * timestep, self.p['a'][j][i])

            for k in range(n_u):
                for j in range(n_bk):
                    formula += ' {}.move({}s)*({})\n+'.format(uf_name[k], (j + 1) * timestep, self.p['b'][i][j][k])

            formula += str(self.p['c'][i])

            formula_list.append({
                'Name': '{} arx model'.format(y_name[i]),
                'Type': 'CalculatedSignal',
                'Description': 'ARX Model {} Formula '.format(y_name[i]),
                'Formula': formula,
                'Formula Parameters': formula_dic
            })
            
        if 'MA' in self.model_struct:
            try:
                tags = spy.push(data=signal_df, metadata=DataFrame(formula_list), workbook=kwargs['workbook_id'], worksheet=kwargs['worksheet_name'], status=spy.Status(quiet=True),
                                quiet=True)
            except:
                tags = spy.push(data=signal_df, metadata=DataFrame(formula_list), workbook=kwargs['workbook_id'], worksheet=kwargs['worksheet_name'], status=spy.Status(quiet=True))
            
            for tag in range(len(yf_name)):
                formula_dic[yf_name[tag]+'am'] = tags[tags['Name'] == y_name[tag]+' arx model']
            
            for i in range(n_y):                
                formula = ''
                formula += '{} - {}'.format(yf_name[i], yf_name[i]+'am')
            
                formula_list.append({'Name': '{} arx error'.format(y_name[i]),
                                     'Type': 'CalculatedSignal',
                                     'Description': 'ARX Model {} Error Formula '.format(y_name[i]),
                                     'Formula': formula,
                                     'Formula Parameters': formula_dic
                })
                
            try:
                tags = spy.push(data=signal_df, metadata=DataFrame(formula_list), workbook=kwargs['workbook_id'], worksheet=kwargs['worksheet_name'], status=spy.Status(quiet=True),
                                quiet=True)
            except:
                tags = spy.push(data=signal_df, metadata=DataFrame(formula_list), workbook=kwargs['workbook_id'], worksheet=kwargs['worksheet_name'], status=spy.Status(quiet=True))
            
            # Moving Average Order
            n_v = self.nv
            
            for i in range(n_y):
                formula_dic[yf_name[i]+'ae'] = tags[tags['Name'] == y_name[tag]+' arx error']
                formula = ' {}\n+'.format(yf_name[i]+'am')
                for j in range(n_v):
                    formula += ' {}.move({}s)*({})\n+'.format(yf_name[i]+'ae', (j+1) * timestep, self.p['d'][i][j][0])
                    
                formula = formula + ' 0.0'
                formula_list.append({'Name': '{} arimax model'.format(y_name[i]),
                                     'Type': 'CalculatedSignal',
                                     'Description': 'ARIMAX Model {} Formula '.format(y_name[i]),
                                     'Formula': formula,
                                     'Formula Parameters': formula_dic
                })
            

        self.formula = DataFrame(formula_list)

# from pandas import read_csv
# import matplotlib.pyplot as plt
#
#
# x_df = read_csv('../../signal_df.csv')
# x_df.set_index('Time', drop=True, inplace=True)
# x_df = x_df.iloc[50:]
# arx = ARX()
#
# arx.mv = ['F_cw']
# arx.cv = ['T1', 'T2']
# arx.identify(x_df)
# yp = arx.forecast(x_df)
# plt.plot(yp)
# plt.show()



from .app_sheet import AppSheet, ARXAppSheet, SSAppSheet, NNAppSheet
from .app_bar import AppBar
from .panels import LeftPanel, ARXPanel, SSPanel, NNPanel
from .figure_table import FigureTable
from .utils import *

from .setup import Setup
from .transfer_matrix import TransferMatrix
from .transfer_card_setup import TransferCardSetup
from .transfer_card_result import TransferCardResult
from .transfer_chip import TransferChip



import ipyvuetify as v

import plotly.graph_objects as go

from copy import deepcopy

class TransferCardResult(v.Dialog):
        
    def __init__(self,
                 dense=True,
                 class_='ma-4 pa-2',
                 style_='corner-radius:10px',
                 **kwargs):
                
        super().__init__(dense=dense,
                         style_=style_,
                         max_width='800px',
                         v_model=False,
                         overlay_opacity=0,
                         **kwargs)
        
        self.on_event('keydown.stop', lambda *args: None)

        self.model_equation = v.CardTitle(children=[], class_='ma-auto pa-auto pb-0 pt-0', align='top', dense=True)

        self.done_btn = v.Btn(children=['Done'], dark=True, color='success', class_='ma-0 pa-0 ml-4 mb-1', dense=True, no_gutters=True)
        
        self.card_switch = v.CardActions(children=[self.done_btn], no_gutters=True, class_='ma-auto pa-auto', dense=True)

        # Figure
        fig = go.FigureWidget()
        fig['layout']['showlegend'] = False
        fig['layout']['xaxis']['rangeslider']['visible'] = True
        fig['layout']['xaxis']['title'] = 'Time'
        fig['layout']['margin']['t'] = 10
        fig['layout']['margin']['b'] = 10
        fig['layout']['margin']['r'] = 10
        fig['layout']['margin']['l'] = 10

        fig['layout']['width'] = 480
        fig['layout']['height'] = 440

        fig['layout']['yaxis']['fixedrange'] = False
        
        self.blank_fig = fig
        self.fig = deepcopy(fig)
        self.fig_item = v.Card(children=[self.fig], style_='width:10%', class_='ma-auto pa-auto pt-0', elevation=0)


        # Model Properties
        self.model_prop_title = v.Text(children=['Model Properties: '], class_='mx-1 my-0 pt-2 pb-1', align='center', dense=True, style_='font-size:14pt; font-weight:bold')
        self.model_prop_item = v.Text(children=['Order: '], class_='ma-1 py-0 pl-4', align='center', dense=True, style_='font-size:12pt')
        
        # Step Information
        self.step_info_title = v.Text(children=['Step Info: '], class_='mx-1 my-0 pt-2 pb-1', align='center', dense=True, style_='font-size:14pt; font-weight:bold')
        self.step_info_item = v.Text(children=[v.Text(children=['Overshoot(%): ']), v.Text(children=['Rise Time: ']), v.Text(children=['Settling Time: '])],
                                     class_='d-flex flex-column ma-1 py-0 pl-4', align='center', dense=True, style_='font-size:13pt')
        
        
        self.model_info = v.Card(children=[self.model_prop_title, self.model_prop_item, self.step_info_title, self.step_info_item], 
                                 class_='d-flex flex-column justify-space-between',
                                 style_='width:35%',
                                 elevation=0)
        
        self.card_item = v.CardActions(children=[self.model_info, v.Spacer(), self.fig], style_='', class_='ma-0 pa-0 ml-2 align-start', elevation=0, disabled=False)

        self.dialog_card = v.Card(class_='pa-2', flat=True)
        self.dialog_card.children = [v.Row(children=[self.model_equation, v.Spacer(), self.card_switch], class_='pa-auto ma-auto mx-0', no_gutters=True, dense=True),
                       v.Divider(class_='mt-auto pt-auto mx-1 mb-2', dense=True),
                       self.card_item]
        self.children = [v.Card(children=[self.dialog_card])]
        
        
    def update_figure(self, df):
        self.fig.data = []
        
        self.fig['layout']['yaxis']['showticklabels'] = True
        step_plot = go.Scatter(x=df.index, y=df, line={'dash': 'solid'}, mode='lines')
        data = [step_plot]
        self.fig.add_traces(data)


import os
import sys
import argparse
import subprocess
from getpass import getpass
from urllib.parse import urlparse
from seeq import sdk, spy

from seeq.spy import _url
from ._copy import copy
import seeq

NB_EXTENSIONS = ['widgetsnbextension', 'plotlywidget', 'ipyvuetify', 'ipyvue']
DEPLOYMENT_FOLDER = 'deployment'
SYSID_NOTEBOOK = "sysid_notebook.ipynb"
DEFAULT_GROUP = ['Everyone']
DEFAULT_USERS = []


def install_app(sdl_url_, *, sort_key='a', permissions_group: list = None, permissions_users: list = None):
    """
    Installs SysID as an Add-on Tool in Seeq Workbench

    Parameters
    ----------
    sdl_url_: str
        URL of the SDL container.
        E.g. `https://my.seeq.com/data-lab/6AB49411-917E-44CC-BA19-5EE0F903100C/`
    sort_key: str, default 'a'
        A string, typically one character letter. The sort_key determines the
        order in which the Add-on Tools are displayed in the tool panel.
    permissions_group: list
        Names of the Seeq groups that will have access to each tool. If None,
        the "Everyone" group will be used by default.
    permissions_users: list
        Names of Seeq users that will have access to each tool. If None, no
        individual users will be given access to the tool.

    Returns
    --------
    -: None
        System Identification will appear as Add-on Tool(s) in Seeq
        Workbench
    """

    permissions_group = permissions_group if permissions_group else DEFAULT_GROUP
    permissions_users = permissions_users if permissions_users else DEFAULT_USERS
    add_on_details = {
        "Name": 'System Identification',
        "Description": "Build mathematical models of dynamic systems from observed input–output data",
        "Icon": "fa fa-star-half-o",
        "Target URL": f'{sdl_url_}/apps/{DEPLOYMENT_FOLDER}/{SYSID_NOTEBOOK}',
        "Link Type": "window",
        "Window Details": "toolbar=0,location=0,left=800,top=400,height=1000,width=1400",
        "Sort Key": sort_key,
        "Reuse Window": True,
        "Groups": permissions_group,
        "Users": permissions_users
        }

    copy(des_folder=DEPLOYMENT_FOLDER, src_folder='deployment_notebook',
         overwrite_folder=False, overwrite_contents=True)
    spy.addons.install(add_on_details, include_workbook_parameters=True, update_tool=True, update_permissions=True)


def install_nbextensions():
    """
    Installs the Jupyter nbextensions required to render the Add-on

    Returns
    -------
    -: None
    """
    for extension in NB_EXTENSIONS:
        subprocess.run(f'jupyter nbextension install --user --py {extension}', cwd=os.path.expanduser('~'), shell=True,
                       check=True)
        subprocess.run(f'jupyter nbextension enable --user --py {extension}', cwd=os.path.expanduser('~'), shell=True,
                       check=True)

def logging_attempts(_user):
    """
    Allows user to re-enter credentials multiple times in the event of
    authentication failure
    Parameters
    ----------
    _user: str
        Seeq username that needs to be authenticated
    Returns
    -------
    -: None
    """
    count = 0
    allowed_attempts = 20
    while count <= allowed_attempts:
        try:
            if _user is None or count >= 1:
                _user = input("\nAccess Key or Username: ")

            passwd = getpass("Access Key Password: ")
            spy.login(username=_user, password=passwd, ignore_ssl_errors=True)
            break
        except (SPyRuntimeError, SPyValueError):
            count += 1
            try_again = "-"
            while try_again != 'yes' and try_again != 'no':
                try_again = input("\nTry again (yes/no)? [yes] ")
                if try_again == '' or try_again.lower() == 'y':
                    try_again = 'yes'
                if try_again.lower() == 'n':
                    try_again = 'no'
            print("-" * 60)
            if try_again.lower() == 'no':
                raise
            if count > allowed_attempts:
                raise RuntimeError("Number of login attempts exceeded")


def cli_interface():
    """ Command line utility to install the SysID Add-on Tool """
    parser = argparse.ArgumentParser(description='Install SysID as a Seeq Add-on Tool')
    parser.add_argument('--nbextensions_only', action='store_true',
                        help='Only installs the nbextensions without installing or updating the Add-on Tools'
                             'links')
    parser.add_argument('--username', type=str,
                        help='Username or Access Key of Seeq admin user installing the tool(s) ')
    parser.add_argument('--seeq_url', type=str, nargs='?',
                        help="Seeq hostname URL with the format https://my.seeq.com/ or https://my.seeq.com:34216")
    parser.add_argument('--users', type=str, nargs='*', default=[],
                        help="List of the Seeq users to will have access to the SysID Add-on Tool,"
                             " default: %(default)s")
    parser.add_argument('--groups', type=str, nargs='*', default=['Everyone'],
                        help="List of the Seeq groups to will have access to the SysID Add-on Tool, "
                             "default: %(default)s")
    return parser.parse_args()


if __name__ == '__main__':

    args = cli_interface()

    if args.nbextensions_only:
        print("\n\nInstalling and enabling nbextensions")
        install_nbextensions()
        sys.exit(0)

    user = args.username
    if user is None:
        user = input("\nAccess Key or Username: ")
    logging_attempts(user)

    seeq_url = args.seeq_url
    if seeq_url is None:
        seeq_url = input(f"\n Seeq base URL [{spy.client.host.split('/api')[0]}]: ")
        if seeq_url == '':
            seeq_url = spy.client.host.split('/api')[0]
    url_parsed = urlparse(seeq_url)
    seeq_url_base = f"{url_parsed.scheme}://{url_parsed.netloc}"

    project_id = spy.utils.get_data_lab_project_id()
    sdl_url = f'{seeq_url_base}/data-lab/{project_id}'
    if project_id is None:
        print("\nThe project ID could not be found. Please provide the SDL project URL with the format "
              "https://my.seeq.com/data-lab/6AB49411-917E-44CC-BA19-5EE0F903100C/\n")
        sdl_url = input("Seeq Data Lab project URL: ")
        project_id = spy.utils.get_data_lab_project_id_from_url(sdl_url)
        if not project_id:
            raise RuntimeError('Could not install "seeq-sysid" because the SDL project ID could not be found')
    sdl_url_sanitized = _url.SeeqURL.parse(sdl_url).url

    print(f"\nThe SysID Tool will be installed on the SDL notebook: {sdl_url_sanitized}\n"
          f"If this is not your intent, you can quit the installation now ")
    print('\n[enter] to continue or type "quit" to exit installation')
    choice = None
    while choice != '' and choice != 'quit':
        choice = input()
        if choice == '':
            print("\n\nInstalling and enabling nbextensions")
            install_nbextensions()
            install_app(sdl_url_sanitized, permissions_group=args.groups, permissions_users=args.users)
        elif choice == 'quit':
            print("\nExited installation")
        else:
            print(f'\nCommand "{choice}" is not valid')
            print('\n[enter] to continue the installation or type "quit" to exit installation')


import plotly.graph_objects as go
import plotly.express as px
from pandas import DataFrame
import ipyvuetify as v
from copy import deepcopy


class DataEditor(v.Card):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fig = None
        self.df0 = DataFrame()
        self.df = DataFrame()
        
        self.cut_btn = v.Btn(children=[v.Icon(children=['mdi-content-cut'], small=True), 'Cut'], class_='mx-1', color='red darken-4', outlined=True)
        self.reset_btn = v.Btn(children=[v.Icon(children=['mdi-refresh'], small=True), 'Reset'], class_='mx-1', color='purple darken-4', outlined=True)
        self.done_btn = v.Btn(children=[v.Icon(children=['mdi-check'], small=True), 'Done'], color='success', class_='mx-1 mr-3', outlined=True)
        
        self.colors = px.colors.qualitative.Plotly
        
        self.cut_btn.on_event('click', self.cut_btn_action)
        self.reset_btn.on_event('click', self.reset_btn_action)

        fig = go.FigureWidget()
        fig.update_layout(
            dragmode='select',
            newshape_line_color='cyan',
        )
        fig.update_traces(marker=dict(size=3))
        fig['layout']['showlegend'] = True
        fig['layout']['xaxis']['rangeslider']['visible'] = False
        fig['layout']['xaxis']['title'] = 'Time'
        fig['layout']['margin']['t'] = 20
        fig['layout']['margin']['b'] = 35
        fig['layout']['margin']['r'] = 0
        fig['layout']['margin']['l'] = 0
        fig['layout']['modebar']['remove'] = ['Autoscale', 'zoomIn2d', 'zoomOut2d', 'toImage', 'lasso']
        
        fig['layout']['yaxis']['title'] = 'Value'
        fig['layout']['yaxis']['fixedrange'] = False
        
        
        self.blank_fig = fig
        self.fig = deepcopy(fig)
        
        self.class_='ma-5 pa-5 my-0'
        self.flat = True
        # self.width='60%'
        self.shaped=True
        self.children = [v.Row(children=[self.reset_btn, self.cut_btn, self.done_btn], class_='justify-end'), fig]

    def update_figure(self, df0: DataFrame):
        self.fig = deepcopy(self.blank_fig)
        self.df = df0.copy()
        
        fig_list = []
        
        for (tag_id, tag) in enumerate(self.df.columns):
            self.fig.add_trace(go.Scatter(x=self.df[tag].dropna().index, y=self.df[tag].dropna(), name=tag, mode='lines+markers',
                                          line=dict(color='rgba' + str(self.hex_to_rgba(h=self.colors[tag_id % 10],alpha=0.8))),
                                          selected_marker_color='black'))
            
        self.fig.update_traces(marker=dict(size=4))

        self.children = [v.Row(children=[self.reset_btn, self.cut_btn, self.done_btn], class_='justify-end'), self.fig]


    def cut_btn_action(self, item, *args):
        try:
            mins = []
            maxs = []
            for scat in self.fig.data:
                if scat.selectedpoints != ():
                    mins.append(min(scat.selectedpoints))
                    maxs.append(max(scat.selectedpoints))

            lb = min(mins)
            ub = max(maxs)

            self.df = self.df.drop(self.df.index[lb:ub+1])
            
            self.update_figure(self.df)
        except:
            pass
        
    def reset_btn_action(self, item, *args):
        self.df = self.df0.copy()
        self.update_figure(self.df)
        
    def hex_to_rgba(self, h, alpha):
        return tuple([int(h.lstrip('#')[i:i+2], 16) for i in (0, 2, 4)] + [alpha])
    
    def set_data(self, df0):
        self.df0 = df0.copy()



import ipyvuetify as v

import plotly.graph_objects as go


class TransferCardSetup(v.Card):
    
    item_id = 0
    
    def __init__(self,
                 mv_name=None,
                 cv_name=None,
                 dense=True,
                 elevation=3,
                 class_='ma-0 px-4 pt-1 pb-3',
                 style_='max-width:600px; corner-radius:10px',
                 shaped=False,
                 flat=False,
                 **kwargs):
        
        if mv_name:
            self.mv_name = mv_name
        else:
            self.mv_name = 'MV'+str(TransferCardSetup.item_id) 
            
        if cv_name:
            self.cv_name = cv_name
        else:
            self.cv_name = 'CV'+str(TransferCardSetup.item_id) 
                
        super().__init__(dense=dense,
                         elevation=elevation,
                         class_=class_,
                         style_=style_,
                         shaped=shaped,
                         flat=flat,
                         **kwargs)

        self.card_name = v.CardTitle(children=[self.mv_name], class_='ma-auto pa-auto pb-0 pt-0', align='top', dense=True)

        self.done_btn = v.Btn(children=['Done'], dark=True, color='success', class_='ma-0 pa-0 ml-4 mb-1', dense=True, no_gutters=True)
        
        signal_switch_list = ['Off', 'On']
        self.signal_on_off_switch = v.Switch(inset=True, class_='ma-0 pa-0 mt-4', v_model=True, label='On', color='success', dark=False, no_gutters=True, dense=True, ripple=False, height='0px')
        self.signal_on_off_switch.label = signal_switch_list[self.signal_on_off_switch.v_model]

        self.card_switch = v.CardActions(children=[self.signal_on_off_switch], no_gutters=True, class_='ma-auto pa-auto', dense=True)


        # Figure
        fig = go.FigureWidget()
        fig['layout']['showlegend'] = False
        fig['layout']['margin']['t'] = 10
        fig['layout']['margin']['b'] = 10
        fig['layout']['margin']['r'] = 0
        fig['layout']['margin']['l'] = 10

        fig['layout']['width'] = 280
        fig['layout']['height'] = 240

        fig['layout']['yaxis']['fixedrange'] = False
        
        self.fig = fig
        self.fig_item = v.Card(children=[self.fig], style_='width:10%', class_='ma-auto pa-auto pt-0', elevation=0)


        # Radio Buttons
        self.order_group = v.RadioGroup(label='Order:', v_model=1, class_='mx-1 my-0 py-0', align='center', dense=True)
        self.radio_fopdt = v.Radio(label='FOPDT', v_model=1, color='blue darken-4', ripple=False, class_='my-0', dense=True)
        self.radio_sopdt = v.Radio(label='SOPDT', v_model=2, color='blue darken-4', ripple=False, class_='my-0', dense=True)
        self.order_group.children = [self.radio_fopdt, self.radio_sopdt]


        # checkboxes
        self.ramp_checkbox = v.Checkbox(label='Ramp', color='blue darken-4', class_='ma-auto pa-0 ml-0', dense=True, ripple=False, v_model=False, height='9px')
        self.gain_checkbox = v.Checkbox(label='Gain', color='blue darken-4', class_='ma-auto pa-0 ml-0', dense=True, ripple=False, v_model=True, height='9px')
        self.deadtime_checkbox = v.Checkbox(label='Deadtime', color='blue darken-4', class_='ma-auto pa-0 ml-0', dense=True, ripple=False, v_model=False, height='9px')
        
        self.options_group = v.Card(children=[self.gain_checkbox, self.deadtime_checkbox, self.ramp_checkbox],
                         no_gutters=True,
                         dense=True, 
                         class_='ma-auto pa-auto mt-1 ml-2 mb-0 pb-0 mr-6 pr-3',
                         elevation=0)
        
        # Constraints
        self.gain_lb = v.TextField(label='lb', v_model='', class_='ml-3 mt-0 pt-0', dense=True, style_='max-width:20%')
        self.gain_ub = v.TextField(label='ub', v_model='', class_='ml-3 mt-0 pt-0', dense=True, style_='max-width:20%')
        self.gain_limit = v.Card(children=['Gain:', v.Spacer(), self.gain_lb, self.gain_ub],
                                 class_='d-flex d-row justify-right align-center mr-3 mt-0 pt-0',
                                 style_='font-size:12pt',
                                 elevation=0, flat=True)

        self.tau_lb = v.TextField(label='lb', v_model='0', class_='ml-3 mt-0 pt-0', dense=True, style_='max-width:20%')
        self.tau_ub = v.TextField(label='ub', v_model='', class_='ml-3 mt-0 pt-0', dense=True, style_='max-width:20%')
        self.tau_limit = v.Card(children=['Time Constant:', v.Spacer(), self.tau_lb, self.tau_ub],
                                class_='d-flex d-row justify-right align-center mr-3 mt-0 pt-0',
                                style_='font-size:11pt',
                                elevation=0, flat=True)
        
        self.deadtime_lb = v.TextField(label='lb', v_model='0', class_='ml-3 mt-0 pt-0', dense=True, style_='max-width:20%', disabled=True)
        self.deadtime_ub = v.TextField(label='ub', v_model='', class_='ml-3 mt-0 pt-0', dense=True, style_='max-width:20%', disabled=True)
        self.deadtime_limit = v.Card(children=['Deadtime:', v.Spacer(), self.deadtime_lb, self.deadtime_ub],
                                     class_='d-flex d-row justify-right align-center mr-3 mt-0 pt-0',
                                     style_='font-size:11pt', elevation=0, flat=True)

        self.constraints_card = v.Card(class_='ma-auto pa-auto mt-0', elevation=0, flat=True, dense=True)
        self.constraints_card.children = [self.gain_limit, self.tau_limit, self.deadtime_limit]
        self.tf_struct_options = v.Card(children=[self.order_group, v.Spacer(), self.options_group],
                                 no_gutters=True,
                                 dense=True, 
                                 class_='d-flex flex-row ma-auto pa-0 mt-3',
                                 elevation=0)
        
        self.tf_options = v.Card(children=[self.tf_struct_options, self.constraints_card], class_='', elevation=0)

        self.card_item = v.CardActions(children=[self.tf_options, v.Spacer(), self.fig], style_='', class_='ma-0 pa-0 ml-2', elevation=0, disabled=False)

        self.children=[v.Row(children=[self.card_name, v.Spacer(), self.card_switch],
                             class_='pa-auto ma-auto mx-0', 
                             no_gutters=True, dense=True),
                       v.Divider(class_='mt-auto pt-auto mx-1 mb-2', dense=True),
                       self.card_item]
        
        TransferCardSetup.item_id += 1
        
    def update_figure(self, df):
        self.fig['layout']['yaxis2'] = {'overlaying': 'y', 'side': 'right'}
        self.fig['layout']['yaxis']['showticklabels'] = True
        self.fig['layout']['yaxis2']['showticklabels'] = True
        cv_plot = go.Scatter(x=df.index, y=df[self.cv_name], name=self.cv_name, yaxis='y1')
        mv_plot = go.Scatter(x=df.index, y=df[self.mv_name], name=self.mv_name, yaxis='y2', line={'dash': 'dash'})
        data = [cv_plot, mv_plot]
        self.fig.add_traces(data)

        
        
    # Reset Item ID
    @staticmethod
    def reset():
        TransferCardSetup.item_id = 0



"""
This file contains some functions (it could be classes) to perform backend calculations
"""

from urllib.parse import parse_qs, unquote, urlparse
from IPython.display import clear_output

from pandas import DataFrame
from seeq import spy


def pull_signals(url, grid='auto'):
    worksheet = spy.utils.get_analysis_worksheet_from_url(url)
    start = worksheet.display_range['Start']
    end = worksheet.display_range['End']

    try:
        search_df = spy.search(url, estimate_sample_period=worksheet.display_range, quiet=True,
                               status=spy.Status(quiet=True))
    except:
        search_df = spy.search(url, estimate_sample_period=worksheet.display_range, status=spy.Status(quiet=True))

    capsules_list = search_df[search_df['Type'].str.contains('CalculatedCondition')]['Name'].to_list()
    signal_list = search_df[search_df['Type'].str.contains('Signal')]['Name'].to_list()

    if search_df.empty:
        return DataFrame(), DataFrame(), DataFrame()
    search_all_df = search_df[search_df['Type'].str.contains('al')]

    try:
        all_df = spy.pull(search_all_df, start=start, end=end, grid=grid, header='Name', quiet=True,
                          status=spy.Status(quiet=True))
    except:
        all_df = spy.pull(search_all_df, start=start, end=end, grid=grid, header='Name', status=spy.Status(quiet=True))
        
    all_df = all_df[search_all_df['Name']]
    
    if all_df.empty:
        return DataFrame(), DataFrame(), DataFrame()
    
    # if hasattr(all_df, 'spy') and hasattr(all_df.spy, 'query_df'):
    #     all_df.columns = all_df.spy.query_df['Name']
    # elif hasattr(all_df, 'query_df'):
    #     all_df.columns = all_df.query_df['Name']
    # else:
    #     raise AttributeError(
    #         "A call to `spy.pull` was successful but the response object does not contain the `spy.query_df` property "
    #         "required for `seeq_sysid")

    all_df.dropna(inplace=True)
    signal_df = all_df[signal_list]
    if signal_df.empty:
        return DataFrame(), DataFrame(), DataFrame()
    capsule_df = all_df[capsules_list]
    if capsule_df.empty:
        return signal_df, DataFrame(), search_df
    signal_df.columns = signal_list
    capsule_df.columns = capsules_list

    clear_output()

    return signal_df, capsule_df, search_df


def parse_url(url):
    unquoted_url = unquote(url)
    return urlparse(unquoted_url)


def get_worksheet_url(jupyter_notebook_url):
    parsed = parse_url(jupyter_notebook_url)
    params = parse_qs(parsed.query)
    return f"{parsed.scheme}://{parsed.netloc}/workbook/{params['workbookId'][0]}/worksheet/{params['worksheetId'][0]}"

def get_workbook_worksheet_workstep_ids(url):
    parsed = parse_url(url)
    params = parse_qs(parsed.query)
    workbook_id = None
    worksheet_id = None
    workstep_id = None
    if 'workbookId' in params:
        workbook_id = params['workbookId'][0]
    if 'worksheetId' in params:
        worksheet_id = params['worksheetId'][0]
    if 'workstepId' in params:
        workstep_id = params['workstepId'][0]
    return workbook_id, worksheet_id, workstep_id


def push_signal(df, workbook_id, worksheet_name):
    try:
        spy.push(df, workbook=workbook_id, worksheet=worksheet_name, status=spy.Status(quiet=True), quiet=True)
    except:
        spy.push(df, workbook=workbook_id, worksheet=worksheet_name, status=spy.Status(quiet=True))


def push_formula(df, formula, workbook_id, worksheet_name):
    try:
        spy.push(data=df, metadata=formula, workbook=workbook_id, worksheet=worksheet_name, status=spy.Status(quiet=True),
                 quiet=True)
    except:
        spy.push(data=df, metadata=formula, workbook=workbook_id, worksheet=worksheet_name, status=spy.Status(quiet=True))

def hide_temp_formula(workbook_id, worksheet_name):
    try:
        workbook_df = spy.workbooks.search({'ID':workbook_id}, quiet=True, status=spy.Status(quiet=True))
        workbook = spy.workbooks.pull(workbooks_df=workbook_df, quiet=True, status=spy.Status(quiet=True))[0]
    except:
        workbook_df = spy.workbooks.search({'ID':workbook_id}, status=spy.Status(quiet=True))
        workbook = spy.workbooks.pull(workbooks_df=workbook_df, status=spy.Status(quiet=True))[0]
        
    worksheet = workbook.worksheet(worksheet_name)
    items = worksheet.display_items
    
    workbook.worksheet(worksheet_name).display_items = items[~items['Name'].str.contains('arx')]
    
    try:
        spy.workbooks.push(workbooks=workbook, quiet=True, status=spy.Status(quiet=True))
    except:
        spy.workbooks.push(workbooks=workbook, status=spy.Status(quiet=True))



import os
from pathlib import Path
from distutils import dir_util
import shutil

sysid_dir = Path(__file__).resolve().parent


def copy(des_folder=None, src_folder=None, *, overwrite_folder=False, overwrite_contents=False):
    """
    Copies the files on the src_folder to a des_folder in the current working directory.
    This function should be called again with overwrite=True if the SysID package is updated.

    Parameters
    ----------
    des_folder : str
        The destination folder. By default it will be copied to a 'deployment_notebook' folder
         in the current working directory.

    src_folder : str
        Name of the folder that contains the documentation to be copied

    overwrite_folder : bool
        If True, any existing files in the specified folder will be deleted before the documentation is copied in.

    overwrite_contents : bool
        If True, files in the specified folder will be overwriting if necessary when the documentation is copied in. 
    """

    validate_argument_types([
        (des_folder, 'folder', str),
        (src_folder, 'overwrite', str),
        (overwrite_folder, 'overwrite', bool),
        (overwrite_contents, 'overwrite', bool),
    ])
    root_src_dir = os.path.dirname(__file__)
    print(root_src_dir)

    if not des_folder:
        des_folder = 'deployment'
    des_folder_path = Path.cwd().joinpath(des_folder)

    if not src_folder:
        src_folder = 'deployment_notebook'
    src_folder_path = sysid_dir.joinpath(src_folder)

    if des_folder_path.exists():
        if not overwrite_folder and not overwrite_contents:
            raise RuntimeError(
                f"The {des_folder_path} directory already exists. If you would like to overwrite it, supply the "
                f"overwrite_folder=True parameter to clean up the folder or overwrite_contents=True to keep the "
                f"current files but overwrite them with the contents of {src_folder_path}. ")
        if overwrite_folder:
            dir_util.remove_tree(str(des_folder_path))
            dir_util.copy_tree(src_folder_path, str(des_folder_path))
            print(f'Copied SysID notebook to {des_folder_path}')
        if overwrite_contents:
            if not os.path.exists(des_folder_path):
                os.makedirs(des_folder_path)
            for src_dir, dirs, files in os.walk(src_folder_path):
                for file_ in files:
                    src_file = os.path.join(src_dir, file_)
                    dst_file = os.path.join(des_folder_path, file_)
                    if os.path.exists(dst_file):
                        # in case of the src and dst are the same file
                        if os.path.samefile(src_file, dst_file):
                            continue
                        os.remove(dst_file)
                    shutil.copy(src_file, des_folder_path)

    else:
        dir_util.copy_tree(src_folder_path, str(des_folder_path))
        print(f'Copied SysID notebook to {des_folder_path}')


def validate_argument_types(expected_types):
    for _value, _name, _types in expected_types:
        if _value is None:
            continue

        if not isinstance(_value, _types):
            if isinstance(_types, tuple):
                acceptable_types = ' or '.join([_t.__name__ for _t in _types])
            else:
                acceptable_types = _types.__name__

            raise TypeError("Argument '%s' should be type %s, but is type %s" % (_name, acceptable_types,
                                                                                 type(_value).__name__))

    return {_name: _value for _value, _name, _types in expected_types}


import ipyvuetify as v
from ipywidgets import HTML


class Setup(v.Card):  
    colors = {
        'app_bar': '#007960',
        'controls_background': '#F6F6F6',
        'visualization_background': '#FFFFFF',
        'seeq_primary': '#007960',
    }

    additional_styles = HTML("""
        <style>
        .background_box { background-color:#007960 !important; } 
        .js-plotly-plot .plotly .modebar-btn[data-title="Produced with Plotly"] {display: none;}
        .vuetify-styles .theme--light.v-list-item .v-list-item__action-text, 
        .vuetify-styles .theme--light.v-list-item .v-list-item__subtitle {color: #212529;}
        .vuetify-styles .theme--light.v-list-item:not(.v-list-item--active):not(.v-list-item--disabled) 
        {color: #007960 !important;}
        .vuetify-styles .v-label {font-size: 14px;}
        .vuetify-styles .v-application {font-family: "Source Sans Pro","Helvetica Neue",Helvetica,Arial,sans-serif;}
        </style>""")
    v.theme.themes.light.success = '#007960'
    v.theme.themes.light.primary = '#007960'
    
    
    def __init__(self):
        super().__init__()
        self.class_ = 'd-flex flex-column ma-0 pa-0 align-content-center flex-wrap justify-space-around'
        self.style_ = 'width:100%; min-height:750px; height:100%; background-color:#1D376C'

        
        # Select MV and CV
        self.mv_title = v.Text(children=['Manipulated Variables (MV)'], style_='font-size:12pt; font-weight:bold')
        self.mv_select = v.Select(tag='Manipulated Variables',
                                  v_model=[],
                                  items=[],
                                  color=self.colors['seeq_primary'],
                                  item_color=self.colors['seeq_primary'],
                                  dense=True,
                                  outlined=False,
                                  class_='d-flex justify-center mt-1',
                                  style_='width: 100%; font-size:14px',
                                  filled=True,
                                  background_color='#EDF4FF',
                                  placeholder='Select',
                                  multiple=True,
                                  clearable=True,
                                  solo=True)
        
        self.cv_title = v.Text(children=['Measured Variables (CV)'], style_='font-size:12pt; font-weight:bold')
        self.cv_select = v.Select(tag='Measured Variables',
                          v_model=[],
                          items=[],
                          color=self.colors['seeq_primary'],
                          item_color=self.colors['seeq_primary'],
                          dense=True,
                          class_='d-flex justify-center mt-1',
                          style_='width: 100%; font-size:14px',
                          outlined=False,
                          filled=True,
                          background_color='#EDF4FF',
                          placeholder='Select',
                          multiple=True,
                          clearable=True,
                          solo=True)
        
        
        
        # Select Train Condition and Validation Condition
        self.train_condition_title = v.Text(children=['Training Conditions'], style_='font-size:12pt; font-weight:bold')
        self.train_condition_select = v.Select(tag='Training Conditions',
                                               v_model=[],
                                               items=[],
                                               color=self.colors['seeq_primary'],
                                               item_color=self.colors['seeq_primary'],
                                               dense=True,
                                               outlined=False,
                                               class_='d-flex justify-center mt-1',
                                               style_='width: 100%; font-size:14px',
                                               filled=True,
                                               background_color='#EDF4FF',
                                               placeholder='Select',
                                               multiple=True,
                                               clearable=True,
                                               solo=True)
        
        self.valid_condition_title = v.Text(children=['Validation Conditions'], style_='font-size:12pt; font-weight:bold')
        self.valid_condition_select = v.Select(tag='Validation Conditions',
                                               v_model=[],
                                               items=[],
                                               color=self.colors['seeq_primary'],
                                               item_color=self.colors['seeq_primary'],
                                               dense=True,
                                               class_='d-flex justify-center mt-1',
                                               style_='width: 100%; font-size:14px',
                                               outlined=False,
                                               filled=True,
                                               background_color='#EDF4FF',
                                               placeholder='Select',
                                               multiple=True,
                                               clearable=True,
                                               solo=True)
        
        
        self.select_card = v.Card(children=[self.mv_title, self.mv_select,
                                            self.cv_title, self.cv_select,
                                            self.train_condition_title, self.train_condition_select,
                                            self.valid_condition_title, self.valid_condition_select],
                                  class_='ma-2 pa-2 justify-center', 
                                  style_='width:98%', 
                                  flat=True, elevation=0)
        
        # Navigation Buttons
        self.next_btn = v.Btn(children=['Next >'], 
                              class_='d-flex ma-2 justify-end white--text',
                              bold=True,
                              color='#1D376C',
                              style_='font-weight:bold; width:80px; text-transform:none;',
                              dark=False,
                              disabled=True)
        self.next_btn_container = v.Layout(children=[v.Spacer(), self.next_btn], class_='d-flex flex-column justify-end align-end')
        
        self.items_card = v.Card(children=[self.select_card], class_='pa-2 ma-2 align-center', style_='width:30%; max-height:450px')
        
        self.signal = []
        self.capsules = []
        
        self.children=[self.items_card]
        
    # Functions
    def get_data(self):
        return (self.mv_select.v_model, self.cv_select.v_model)
    
    def set_data(self, signals: list=[], capsules: list=[]):
        self.mv_select.items = signals
        self.cv_select.items = signals
        
        self.train_condition_select.items = capsules
        self.valid_condition_select.items = capsules

        self.signal = signals
        self.capsules = capsules



import ipyvuetify as v
import ipywidgets as widgets
from pathlib import Path
from .utils import add_tooltip
from .data_editor import DataEditor

import io
from pandas import read_csv


class HamburgerMenu(v.Menu):
    """
    Create a Hamburger Menu in the right corner of App Bar
    """

    def __init__(self, server=True, **kwargs):
        self.server = server
        # Create menu icon button
        self.hamburger_button = v.AppBarNavIcon(v_on='menuData.on', class_='align-center mb-2 mt-3 ml-3 mr-0')
        
        
        # load worksheet option
        self.load_worksheet = v.ListItem(value='open',
                                         children=[v.ListItemAction(class_='mr-2 ml-0',
                                                                       children=[v.Icon(color='#212529',
                                                                                        children=['mdi-folder-search-outline'])]),
                                                      v.ListItemActionText(children=[f'Open Worksheet'])
                                                      ])
        
        # load worksheet option
        self.load_data_editor = v.ListItem(value='open',
                                         children=[v.ListItemAction(class_='mr-2 ml-0',
                                                                       children=[v.Icon(color='#212529',
                                                                                        children=['mdi-folder-edit-outline'])]),
                                                      v.ListItemActionText(children=[f'Data Editor'])
                                                      ])
        

        # Feedback option
        self.feedback_button = v.ListItem(value='help',
                                          ripple=True,
                                          href='https://github.com/BYU-PRISM/Seeq/issues',
                                          children=[v.ListItemAction(class_='mr-2 ml-0',
                                                                     children=[v.Icon(color='#212529',
                                                                                      children=['fa-life-ring'])]),
                                                    v.ListItemActionText(children=[f'Send Support Request'])
                                                    ])
        # GitHub page
        self.user_guide_button = v.ListItem(value='tutorial',
                                            ripple=True,
                                            href='https://github.com/BYU-PRISM/Seeq',
                                            target='_blank',
                                            children=[v.ListItemAction(class_='mr-2 ml-0',
                                                                       children=[v.Icon(color='#212529',
                                                                                        children=['mdi-help-box'])]),
                                                      v.ListItemActionText(children=[f'User Guide'])
                                                      ])
        
        
        
        # Open Worksheet Dialog & Events
        self.load_worksheet.on_event('click', self.load_worksheet_action)
        
        # Open Data Editor
        self.load_data_editor.on_event('click', self.load_data_editor_action)

        self.close_url_dialog_btn = v.Btn(children=['CLOSE'], color='#007960', text=True)
        self.close_url_dialog_btn.on_event('click', self.close_url_action)

        self.ok_url_dialog_btn = v.Btn(children=['OK'], color='#007960', text=True, loading=False)
        self.ok_url_dialog_btn.on_event('click', self.ok_url_action)
        # Create a text field for the worksheet url
        self.worksheet_url = v.TextField(v_model='', placeholder='Worksheet URL', class_='mx-4', color='#007960',
                                         clearable=True)
        self.worksheet_url.on_event('paste.stop', lambda *args: None)
        
        box_layout = widgets.Layout(display='flex',
                flex_flow='column',
                align_items='center',
                width='100%')
        self.upload_btn = widgets.FileUpload(layout=widgets.Layout(width='50%'), style={'button_color': '#007960'})
        self.upload_btn.button_style = 'success'
        self.upload_btn.observe(lambda _:self.upload_file(self.upload_btn), 'data')
        self.upload_lay = widgets.HBox(children=[self.upload_btn], layout=box_layout)
        
        self.sl_switch = v.Switch(v_model=False, label='Local', inset=True, class_='ma-0 pa-0 mt-4 ml-2', dense=True, no_gutters=True, ripple=False, height='0px')
        self.sl_switch.hide()
        self.sl_switch.on_event('change', self.import_mode_switch)
        
        self.control_dialog_btn_layout = v.Layout(children=[v.Spacer(), self.close_url_dialog_btn, self.ok_url_dialog_btn])

        self.dialog_card_content = [v.CardTitle(children=['Please Enter a Worksheet URL:', v.Spacer(), self.sl_switch]),
                               self.worksheet_url,
                               self.control_dialog_btn_layout]

        self.dialog_card = v.Card(children=self.dialog_card_content, class_='pa-2 ma-3 my-0', flat=True)

        self.url_dialog = v.Dialog(name='OpneWB',
                                   children=[v.Card(children=[self.dialog_card])],
                                   v_model=False,
                                   max_width='600px')

        self.url_dialog.on_event('keydown.stop', lambda *args: None)
        
        self.data_editor = DataEditor()
        self.data_editor_dialog = v.Dialog(name='OpneDE',
                           children=[v.Card(children=[self.data_editor], shaped=True)],
                           v_model=False,
                           max_width='60%')
        
        self.items = [self.load_worksheet, v.Divider(), self.load_data_editor, v.Divider(), self.feedback_button, v.Divider(), self.user_guide_button, self.url_dialog, self.data_editor_dialog]
        
        self.local_data = None
        
        super().__init__(offset_y=True,
                         offset_x=False,
                         dense=True,
                         right=True,
                         class_='ma-0 pa-0',
                         v_slots=[{
                             'name': 'activator',
                             'variable': 'menuData',
                             'children': self.hamburger_button
                         }],
                         children=[
                             v.List(children=self.items)
                         ],
                         **kwargs)
        
        
    # Load worksheet event functions
    def load_worksheet_action(self, *args):
        self.url_dialog.v_model = True

    def close_url_action(self, *args):
        self.url_dialog.v_model = None

    def ok_url_action(self, *args):
        self.url_dialog.v_model = None
        
    # DataEditor event functions
    def load_data_editor_action(self, *args):
        self.data_editor_dialog.v_model = True
        
    def import_mode_switch(self, item, *args):
        if item.v_model:
            self.dialog_card_content = [v.CardTitle(children=['Please Import CSV file:', v.Spacer(), self.sl_switch]),
                                        self.upload_lay,
                                        self.control_dialog_btn_layout]
        else:
            self.dialog_card_content = [v.CardTitle(children=['Please Enter a Worksheet URL:', v.Spacer(), self.sl_switch]),
                                        self.worksheet_url,
                                        self.control_dialog_btn_layout]

        self.dialog_card.children = self.dialog_card_content
            
    def upload_file(self, x):
        try:
            uploaded_file = x.data[0]
            self.local_data = read_csv(io.BytesIO(uploaded_file), index_col='Time')
        except:
            print('csv read error! Make sure that DateTime data are in the "Time" column.')

class AppBar(v.Card):
    """
    Create app bar for the Add-on. This app bar includes icon, tabs, load worksheet button and, hamburger menu.
    """

    def __init__(self, tabs=None, items=None, title_list=None, logo_name=Path(__file__).parent / '../data/seeq_logo.png', *args,
                 **kwargs):
        """
        Args:
            tabs : list of tabs
            items : list of items (app sheet)
            logo_name : App icon directory
        """
        super().__init__(flat=True, elevation=0, shadow=False, *args, **kwargs)
            
        # read add-on logo (.png)
        file = open(logo_name, 'rb')
        image = file.read()
        self.image = widgets.Image(value=image, format='png', align='center')
        file.close()

        self.logo = [v.Img(children=[self.image], class_='align-center pl-4')]
        
        self.title = v.Text(children=[''], style_='font-size:18pt ; font-weight:bold', class_='d-flex justify-end align-center')
        
        self.ham_menu = HamburgerMenu()
        
        self.next_btn = v.Btn(children=['Next >'], 
                  class_='ma-2 mt-3 mr-3 white--text',
                  bold=True,
                  color='#1D376C',
                  style_='font-weight:bold; width:80px; text-transform:none',
                  dark=False)

        self.back_btn = v.Btn(children=['< Back'], 
                  class_='ma-2 mt-3 ml-3',
                  bold=True,
                  color='#1D376C',
                  style_='font-weight:bold; width:80px; text-transform:none; background-color:none',
                  outlined=True,
                  dark=True,
                  disabled=False)
        
        self.push_btn = v.Btn(children=['Push'],
                              class_='ma-2 mt-3 mr-3 white--text',
                              bold=True,
                              color='#1D376C',
                              style_='font-weight:bold; width:80px; text-transform:none',
                              dark=False)
        
        self.nav_btn = v.Layout(children=[self.back_btn, self.next_btn], class_='d-flex justify-end align-center ma-2')

        if items is None:
            items = []
        if tabs is None:
            tabs = []
        if title_list is None:
            self.nav_btn = v.Layout()
            self.title_list = []
        else:
            self.title_list = title_list
    
        self.tabs = v.Tabs(children=[self.ham_menu] + self.logo + tabs + [self.title] + items + [self.nav_btn],
                     color='#1d376c',
                     class_='align-center pt-0',
                     slider_size='4',
                     background_color='grey lighten-4',
                     hide_slider=True,
                     height=60,
                     style_='width:98%; border-radius:12px')

        self.children = [self.tabs]




import ipyvuetify as v

from .table_card import TableCard, TableItem, TableHeader
from .figure_card import FigureCard

from pandas import DataFrame


# Visualization Class (Figures+Table)
class FigureTable(v.Card):
    def __init__(self, 
                 class_ = 'd-flex flex-column justify-start align-center ml-1 mr-1',
                 style_ = 'width:100%; height:100%',
                 *args, **kwargs):

        super().__init__(class_=class_,
                         style_=style_,
                         flat=True,
                         *args,
                         **kwargs)

        self.figure = FigureCard()
        self.table = TableCard()

        self.children = [self.figure, self.table]

        self.table.table_header.checkbox_header.observe(self.header_checkbox_action, names=['v_model'])

    def header_checkbox_action(self, item, *_):
        all_value = item.owner.v_model
        for tag in self.table.items_list:
            tag.checkbox_item.v_model = all_value

    def create(self, train_df=DataFrame(), validation_df=DataFrame(), units=None):
        TableItem().reset()
        self.table.table_header.checkbox_header.v_model = False

        if not units:
            units = {}
            for tag in train_df.columns:
                units[tag] = '-'

        # create plots
        self.figure.plot(train_df=train_df, validation_df=validation_df)

        # create table
        items_list = []
        for name in train_df.columns:
            temp_item = TableItem(name=name, unit=units[name])
            temp_item.checkbox_item.observe(self.update_visibility, names=['v_model'])
            temp_item.style_item.on_event('change', self.update_line_style)
            items_list += [temp_item]

        self.table.create_table(items_list)

    def update_visibility(self, item, *_):
        idx = int(item.owner.model_id)
        status = item.owner.v_model

        self.figure.update(idx=idx, status=status)

    def update_line_style(self, item, *_):
        idx = item.idx
        line_style = item.get_style()
        self.figure.update(idx=idx, line_style=line_style)



import pandas as pd
from numpy import zeros, ones, array, empty, arange, exp, pi, sqrt, nan
from pandas import DataFrame, read_csv

from gekko import GEKKO
from gekko.gekko import EquationObj

import matplotlib.pyplot as plt


class TransferOption:
    idx = 0

    def __init__(self, 
                 order=1,
                 no_gain=0,
                 no_ramp=1,
                 is_dt=1,
                 gain_lb=None,
                 gain_ub=None,
                 tau_lb=0,
                 tau_ub=None,
                 theta_lb=0,
                 theta_ub=None,
                 name=None):
        
        self.name = name
        self.id = None

        self.order = order
        self.no_gain = no_gain
        self.no_ramp = no_ramp
        self.is_zeta = 1
        self.is_dt = is_dt

        self.gain = None
        self.gain_lb = gain_lb
        self.gain_ub = gain_ub
        
        self.tau = None
        self.tau_lb = tau_lb
        self.tau_ub = tau_ub
        
        self.theta = None
        self.theta_lb = theta_lb
        self.theta_ub = theta_ub

        self.tau_1 = None
        self.tau_2 = None
        self.zeta = None
        
        self.ts = None
        self.tr = None
        self.os = None

        self.num = None
        self.den = None
        
    def calc_step_info(self, t, yout):
        yout = abs(yout)
        
        if self.order == 1:
            if self.tau:
                self.ts = 4*self.tau
                self.tr = 2.2*self.tau

            self.os = 0.0
        elif self.order == 2:
            if self.zeta == 0:
                self.ts = 1e20
            else:
                self.ts = 3*self.tau/self.zeta
            self.tr = 1.8*self.tau
            if not self.zeta:
                self.os = 100
            elif self.zeta >= 1:
                self.os = 0
            elif self.zeta < 0:
                self.os = nan
            else:
                self.os = 100*exp(-pi*self.zeta/sqrt(1-self.zeta**2))

            # try:
            #     self.ts = t[next(len(yout)-i for i in range(2,len(yout)-1) if abs(yout[-i]/yout[-1])<1.02)]-t[0]
            # except:
            #     self.ts = t[-1]
            # self.tr = t[next(i for i in range(0,len(yout)-1) if yout[i]>yout[-1]*.90)]-t[0]
            # self.os = (yout.max()/yout[-1]-1)*100
        
    def get_model_info(self):
        return (self.order, self.gain, self.tau, self.theta, self.zeta)
        
    def get_step_info(self):
        return (self.ts, self.tr, self.os)
    
    @staticmethod
    def add_idx():
        TransferOption.idx += 1
        
    @staticmethod
    def reset_idx():
        TransferOption.idx = 0


class TransferItem(GEKKO):
    def __init__(self):
        super().__init__(remote=False)

        self.mv = None
        self.cv = None
        
        self.dt = None        
        self.u_ss = None

        self.const_den = False
        self.option_dict = {}

        self.gain = None
        self.tau = None
        self.theta = None
        self.zeta = None
        self.tc = None
        self.tt = None
        
        self.tau_max = None
        self.sim_time = None

        self.t = None
        self.um = None
        self.xm = None
        self.dxm_dt = None
        self.ym = None

        self.nu = None
        self.ny = None

        # self.shift = "mean"
        self.label = None
        self.status = None

    @staticmethod
    def empty_array(mode, dim):
        """
        Create empty numpy arrays to contain gekko objects
        mode = 0 -> zeros
        mode = 1 -> ones
        """
        if mode == 0:
            return zeros(shape=dim, dtype=object)
        if mode == 1:
            return ones(shape=dim, dtype=object)

    def build(self):
        """
        Define a gekko object for the problem
        """
        self.mv = list(self.option_dict.keys())
        nu = len(self.mv)

        # define dead time and dummy time variables
        tt = self.Param(0, name='_time_')
        # self.Equation(tt.dt() == 1)
        tc = empty(shape=nu, dtype=object)

        # define gain (k), time constant (tau) and theta as empty arrays
        gain = empty(shape=nu, dtype=object)
        tau = empty(shape=nu, dtype=object)
        theta = empty(shape=nu, dtype=object)
        zeta = empty(shape=nu, dtype=object)

        # define input and output gekko variables
        um = empty(shape=nu, dtype=object)
        xm = empty(shape=nu, dtype=object)
        ym = empty(shape=1, dtype=object)
        dxm_dt = empty(shape=nu, dtype=object)

        ym[0] = self.CV(value=0, lb=None, ub=None, name=self.cv[0])
        # ym[0].fstatus = 1
        output_eq = "%s=" % (ym[0])

        for i in range(nu):
            option_item: TransferOption = self.option_dict[self.mv[i]]

            # initialize dummy time
            tc[i] = self.Var(value=0, lb=None, ub=None, name="_tc_%i" % (i+1))

            # initialize tau values
            if option_item.no_ramp:
                tau[i] = self.FV(value=0.1, lb=option_item.tau_lb, ub=option_item.tau_ub, name="tau%i" % (i+1))
                tau[i].status = 1
            else:
                tau[i] = 1

            # initialize gain values
            if option_item.no_gain:
                gain[i] = 1
            else:
                gain[i] = self.FV(value=0.1, lb=option_item.gain_lb, ub=option_item.gain_ub, name="K%i" % (i+1))
                gain[i].status = 1

            if option_item.is_dt:
                theta[i] = self.FV(value=1, integer=False, lb=option_item.theta_lb, ub=option_item.theta_ub, name="theta%i" % (i+1))
                theta[i].status = 1
            else:
                theta[i] = 0

            if option_item.order == 2:
                zeta[i] = self.FV(value=2, lb=-10, ub=10, name="zeta%i" % (i + 1))
                zeta[i].status = 1

                dxm_dt[i] = self.Var(value=0, lb=None, ub=None, name='d' + 'x%i' % (i + 1) + '/dt')

            um[i] = self.Var(value=0, lb=None, ub=None, name=self.mv[i])
            xm[i] = self.SV(value=0, lb=None, ub=None, name="x%i" % (i+1))
            xm[i].fstatus = 1

        # create the equations
        for i in range(nu):
            option_item: TransferOption = self.option_dict[self.mv[i]]
            if option_item.order == 1:
                state_eq = "%s*$%s" % (tau[i], xm[i]) if option_item.no_ramp else "$%s" % (xm[i])
            elif option_item.order == 2:
                d_state_eq = "%s=$%s" % (dxm_dt[i], xm[i])
                d_state_eq_obj = EquationObj(d_state_eq)
                self._equations.append(d_state_eq_obj)
                state_eq = "%s^2*$%s" % (tau[i], dxm_dt[i]) if option_item.no_ramp else "$%s" % (dxm_dt[i])
                if option_item.no_ramp:
                    state_eq += "+2*%s*%s*$%s" % (tau[i], zeta[i], xm[i]) if option_item.is_zeta else ""
            else:
                state_eq = "Model Order Not Implemented"
            state_eq += "+%s" % (xm[i]) if option_item.no_ramp else ""
            state_eq += "=%s" % (um[i]) if option_item.no_gain else "=%s*%s" % (gain[i], um[i])

            state_eq_obj = EquationObj(state_eq)
            self._equations.append(state_eq_obj)

            # force same denominator for all the transfer functions corresponding to an output
            if self.const_den and i != 0:
                tau[i].fstatus = 0
                additional_eq = "%s=%s" % (tau[0], tau[i])
                additional_eq_obj = EquationObj(additional_eq)
                self._equations.append(additional_eq_obj)
                if option_item.order == 2:
                    zeta[i].fstatus = 0
                    additional_eq = "%s=%s" % (zeta[0], zeta[i])
                    additional_eq_obj = EquationObj(additional_eq)
                    self._equations.append(additional_eq_obj)

            self.Equation(tc[i] == tt - theta[i])

            output_eq += "%s+" % (xm[i])

        output_eq = output_eq[:-1]
        output_eq_obj = EquationObj(output_eq)
        self._equations.append(output_eq_obj)

        # define the object variables
        self.nu = nu
        self.tc = tc
        self.tt = tt
        self.um = um
        self.xm = xm
        self.dxm_dt = dxm_dt
        self.ym = ym
        self.gain = gain
        self.tau = tau
        self.theta = theta
        self.zeta = zeta

    def identify(self, df):
        df = df.copy()     
        
        # t_df = pd.to_datetime(df.index)
        # t = (t_df.values - t_df.values[0]) / 1e9
        
        if (df.index.dtype == 'float64') or (df.index.dtype =='int64'):
            t = df.index.values
        elif df.index.dtype == 'O':
            t_df = pd.to_datetime(df.index)
            t = (t_df.values - t_df.values[0]) / 1e9
        else:
            t_df = pd.to_datetime(df.index)
            t = (t_df.values - t_df.values[0]) / 1e9
        
        self.dt = float(t[1]-t[0])

        u_df = df[self.mv]
        y_df = df[self.cv]

        # self.u_ss = df[self.mv].mean().to_numpy()
        self.u_ss = df[self.mv].iloc[0].to_numpy()
        u_df -= self.u_ss

        u = u_df.to_numpy().T
        y = y_df.to_numpy().T

        self.t = t
        self.tt.value = t
        self.time = self.t

        for i in range(self.nu):
            option_item: TransferOption = self.option_dict[self.mv[i]]

            if not option_item.no_gain:
                self.gain[i].status = 1
                self.gain[i].fstatus = 1

            if option_item.no_ramp:
                self.tau[i].status = 1
                self.tau[i].fstatus = 1

            if option_item.is_dt:
                self.theta[i].status = 1
                self.theta[i].fstatus = 1

            self.tc[i].value = self.t
            self.um[i].value = u[i]

            self.cspline(self.tc[i], self.um[i], self.t, u[i], bound_x=True)

        self.ym[0].value = y[0]
        # self.ym[0].fstatus = 1
        self.ym[0].status = 0

        self.options.Solver = 3
        self.options.IMODE = 5
        self.solve(disp=False)

        self.label = [tag_label + '_tf' for tag_label in y_df.columns]
        
        # save optimal values
        for (i, mv_i) in enumerate(self.mv):
            option_item: TransferOption = self.option_dict[mv_i]
            option_item.num = [1]
            option_item.den = []

            if not option_item.no_gain:
                option_item.gain = self.gain[i][-1]
                option_item.num = [option_item.gain]

            if option_item.no_ramp:
                option_item.tau = self.tau[i][-1]
            else:
                option_item.tau = 1.0

            if option_item.order == 2:
                if option_item.is_zeta:
                    option_item.zeta = self.zeta[i][-1]
                    option_item.den.append(option_item.tau**2)
                    option_item.den.append(2*option_item.tau*option_item.zeta)
                else:
                    option_item.den.append(option_item.tau ** 2)
                    option_item.den.append(0)
            elif option_item.order == 1:
                option_item.den.append(option_item.tau)

            if option_item.is_dt:
                option_item.theta = self.theta[i][-1]
            else:
                option_item.theta = 0

            if option_item.no_ramp:
                option_item.tau = self.tau[i][-1]
                option_item.den.append(1)
            else:
                if option_item.order == 1:
                    option_item.den = [1, 0]
                elif option_item.order == 2:
                    option_item.den = [1, 0, 0]
                option_item.tau = None

        self.status = True

        return self.ym.copy()

    def simulate(self, u_df):
        u_df = u_df.copy()
        
        self._connections = []
        self._objects = []
        self.clear_data()

        if (u_df.index.dtype == 'float64') or (u_df.index.dtype =='int64'):
            t = u_df.index.values
        elif u_df.index.dtype == 'O':
            t_df = pd.to_datetime(u_df.index)
            t = (t_df.values - t_df.values[0]) / 1e9
        else:
            t_df = pd.to_datetime(u_df.index)
            t = (t_df.values - t_df.values[0]) / 1e9

        self.t = t
        self.time = self.t

        u_df -= self.u_ss
        
        u = u_df.to_numpy().T
        
        for i in range(self.nu):
            option_item: TransferOption = self.option_dict[self.mv[i]]

            if not option_item.no_gain:
                self.gain[i].status = 0
                self.gain[i].fstatus = 0
                self.gain[i].value = self.gain[i].value[0]

            if option_item.no_ramp:
                self.tau[i].status = 0
                self.tau[i].fstatus = 0
                self.tau[i].value = self.tau[i].value[0]

            if option_item.is_dt:
                self.theta[i].status = 0
                self.theta[i].fstatus = 0
                self.theta[i].value = self.theta[i].value[0]

            if option_item.is_zeta and option_item.order == 2:
                self.zeta[i].status = 0
                self.zeta[i].fstatus = 0
                self.zeta[i].value = self.zeta[i].value[0]

            if option_item.order == 2:
                self.dxm_dt[i].value = self.ym[0].value

            self.um[i].value = u[i]
            # self.um[i].status = 0

            self.xm[i].value = 0
            self.xm[i].fstatus = 0
            self.tc[i].value = self.t
            # self.tc[i].lower = 0

            self.cspline(self.tc[i], self.um[i], self.t, u[i], bound_x=True)

        self.tt.value = t
        self.ym[0].value = 0
        self.ym[0].status = 1

        self.options.IMODE = 5
        self.solve(disp=False)

        return self.ym.copy()



from .tf_item import TransferItem, TransferOption
from numpy import empty, array, zeros, arange
from pandas import DataFrame
from scipy.signal import TransferFunction
from .arx import create_formula_variable_name


class TF(TransferItem):
    def __init__(self):
        super().__init__()
        self.y_ss = DataFrame()
        self.models = {}
        self.dummy = None
        self.formula = None
        
    def add_model(self, cv_name, model: TransferItem):
        self.models[cv_name] = model
        self.cv.append(cv_name)

    def identify(self, df, cv_name):
        df = df.copy()

        self.y_ss[cv_name] = df[[cv_name]].iloc[0].to_numpy()

        df -= df.iloc[0]
        
        yp_df = DataFrame()

        # self.nu = len(self.mv)
        self.ny = len(self.cv)

        # for cv_i in self.cv:
        model: TransferItem = self.models[cv_name]
        model.build()
        y_i = model.identify(df=df)
        yp_df[cv_name+"_tf"] = array(y_i[0])
        
        yp_df += self.y_ss[cv_name].to_numpy()
        
        yp_df = yp_df.set_index(df.index)
            
        return yp_df

    def predict(self, u_df, cv_name):
        u_df = u_df.copy()
        
        self.y_ss[cv_name] = u_df[[cv_name]].iloc[0].to_numpy()
        
        u_df -= u_df.iloc[0]
        
        yp_df = DataFrame()

        # for cv_i in self.cv:
        model: TransferItem = self.models[cv_name]
        # u_df -= model.u_ss
        y_i = model.simulate(u_df=u_df[model.mv])
        yp_df[cv_name+"_tf"] = array(y_i[0])
            
        yp_df += self.y_ss[cv_name].to_numpy()
        
        yp_df = yp_df.set_index(u_df.index)
        
        return yp_df
    
    def step_response(self, mv_name, cv_name):
        # simulation time = 12 * tau
        # for cv_i in self.cv:
        model: TransferItem = self.models[cv_name]
        
        # for (i, mv_i) in enumerate(model.mv):
        option_item = model.option_dict[mv_name]
        
        tau = option_item.tau if option_item.tau else 1
        
        tf = max(25*tau, 10*model.dt)
        Time = arange(0, tf+1, model.dt)
        sim_ts = len(Time)

        u = zeros((model.nu, sim_ts))
        u[option_item.id][int(9*tau/model.dt):] = 1
        
        u_df = DataFrame(u.T, columns=model.mv)
        # u_df += model.u_ss
        
        u_df.set_index(Time, inplace=True)
        self.dummy = u_df.copy()
        y_i = model.simulate(u_df=u_df)
        
        yp_df = DataFrame()
        # yp_df[cv_name+"_tf"] = array(y_i[0])/abs(option_item.gain)
        yp_df[cv_name+"_tf"] = array(y_i[0])

        # yp_df += self.y_ss
        yp_df.set_index(Time, inplace=True)
        
        option_item.calc_step_info(yp_df.index.to_numpy(), yp_df.to_numpy().T[0])
        
        return yp_df
    
    def get_step_info(self, mv_name, cv_name):
        model = self.models[cv_name]
        option_item = model.option_dict[mv_name]
        return option_item.get_step_info()
    
    def get_model_info(self, mv_name, cv_name):
        model = self.models[cv_name]
        option_item = model.option_dict[mv_name]
        return option_item.get_model_info()

    def set_sim_time(self):
        # set max time constant for each output (CV)
        for cv_i in self.cv:
            model: TransferItem = self.models[cv_i]
            model.tau_max = 1
            
            for mv_i in model.mv:
                option_item = model.option_dict[mv_i]
                if option_item.tau:
                    if option_item.tau > model.tau_max:
                        model.tau_max = option_item.tau
            
            model.sim_time = 4 * model.tau_max

    def create_formula(self, df, tags):
        y_name = self.cv

        yf_name = create_formula_variable_name(y_name)

        formula_dic = {}
        for tag in range(len(yf_name)):
            formula_dic[yf_name[tag]] = tags[tags['Name'] == y_name[tag]]

        formula_list = []

        for cv_i in range(len(self.cv)):
            cv_name = y_name[cv_i]
            y_ss = df[cv_name].iloc[0]
            y_opt = self.models[cv_name]

            u_name = y_opt.mv
            uf_name = create_formula_variable_name(u_name)
            formula = ''
            for tag in range(len(uf_name)):
                formula_dic[uf_name[tag]] = tags[tags['Name'] == u_name[tag]]

            timestep = y_opt.dt

            for mv_i in range(len(y_opt.mv)):
                mv_name = y_opt.mv[mv_i]
                u_opt = y_opt.option_dict[mv_name]
                u_ss = df[mv_name].iloc[0]

                tf_sys = TransferFunction(u_opt.num, u_opt.den)
                tfd_sys = tf_sys.to_discrete(dt=timestep, method='zoh')
                num = tfd_sys.num
                den = tfd_sys.den

                n_a = len(den)
                n_b = len(num)

                for i in range(1, n_a):
                    formula += ' ({}.move({}s)-{})*({})\n+'.format(yf_name[cv_i], i * timestep, y_ss, -den[i])

                for j in range(n_b):
                    formula += ' ({}.move({}s)-{})*({})\n+'.format(uf_name[mv_i], j * timestep + u_opt.theta, u_ss,
                                                                   num[j])

            formula += str(y_ss)

            formula_list.append({
                'Name': '{} tf'.format(y_name[cv_i]),
                'Type': 'CalculatedSignal',
                'Description': 'TF Model {} Formula '.format(y_name[cv_i]),
                'Formula': formula,
                'Formula Parameters': formula_dic
            })

        self.formula = DataFrame(formula_list)

        return self.formula
